# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Source: https://www.librec.net/datasets.html
"""
import os
from typing import List

from ..utils import cache
from ..data import Reader

from ..utils.download import get_cache_path


def _get_cache_dir():
    cache_dir = get_cache_path("filmtrust")[0]
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    return cache_dir


def load_feedback(reader: Reader = None) -> List:
    """Load the user-item ratings, scale: [0.5,4]

    Parameters
    ----------
    reader: `obj:cornac.data.Reader`, default: None
        Reader object used to read the data.

    Returns
    -------
    data: array-like
        Data in the form of a list of tuples (user, item, rating).
    """
    fpath = cache(
        url="https://static.preferred.ai/cornac/datasets/filmtrust/filmtrust.zip",
        unzip=True,
        cache_dir=_get_cache_dir(),
        relative_path="ratings.txt",
    )
    reader = Reader() if reader is None else reader
    return reader.read(fpath, sep=" ")


def load_trust(reader: Reader = None) -> List:
    """Load the user-user trust information (undirected network)

    Parameters
    ----------
    reader: `obj:cornac.data.Reader`, default: None
        Reader object used to read the data.

    Returns
    -------
    data: array-like
        Data in the form of a list of tuples (user, user, 1).
    """
    fpath = cache(
        url="https://static.preferred.ai/cornac/datasets/filmtrust/filmtrust.zip",
        unzip=True,
        cache_dir=_get_cache_dir(),
        relative_path="trust.txt",
    )
    reader = Reader() if reader is None else reader
    return reader.read(fpath, sep=" ")



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from .common import validate_format
from .common import estimate_batches
from .common import get_rng
from .download import cache
from .fast_dot import fast_dot
from .common import normalize

__all__ = ['validate_format',
           'estimate_batches',
           'get_rng',
           'cache',
           'fast_dot',
           'normalize']



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np


class Modality:
    """Generic class of Modality to extend from
    """

    def __init__(self, **kwargs):
        pass

def fallback_feature(func):
    """Decorator to fallback to `batch_feature` in FeatureModality
    """

    def wrapper(self, *args, **kwargs):
        if self.features is not None:
            ids = args[0] if len(args) > 0 else kwargs['batch_ids']
            return FeatureModality.batch_feature(self, batch_ids=ids)
        else:
            return func(self, *args, **kwargs)

    return wrapper


class FeatureModality(Modality):
    """Modality that contains features in general

    Parameters
    ----------
    features: numpy.ndarray or scipy.sparse.csr_matrix, default = None
        Numpy 2d-array that the row indices are aligned with user/item in `ids`.

    ids: List, default = None
        List of user/item ids that the indices are aligned with `corpus`.
        If None, the indices of provided `features` will be used as `ids`.

    normalized: bool, default = False
        Whether the features will be normalized using min-max normalization.
    """

    def __init__(self, features=None, ids=None, normalized=False, **kwargs):
        super().__init__(**kwargs)
        self.features = features
        self.ids = ids
        self.normalized = normalized

    @property
    def features(self):
        """Return the whole feature matrix
        """
        return self.__features

    @features.setter
    def features(self, input_features):
        if input_features is not None:
            assert len(input_features.shape) == 2
        self.__features = input_features

    @property
    def feature_dim(self):
        """Return the dimensionality of the feature vectors
        """
        return self.features.shape[1]

    def _swap_feature(self, id_map):
        new_feats = np.copy(self.features)
        new_ids = self.ids.copy()
        for old_idx, raw_id in enumerate(self.ids):
            new_idx = id_map.get(raw_id, None)
            if new_idx is None:
                continue
            assert new_idx < new_feats.shape[0]
            new_feats[new_idx] = self.features[old_idx]
            new_ids[new_idx] = raw_id
        self.features = new_feats
        self.ids = new_ids

    def build(self, id_map=None, **kwargs):
        """Build the feature matrix.
        Features will be swapped if the id_map is provided
        """
        if self.features is None:
            return

        if (self.ids is not None) and (id_map is not None):
            self._swap_feature(id_map)

        if self.normalized:
            self.features = self.features - np.min(self.features)
            self.features = self.features / (np.max(self.features) + 1e-10)

        return self

    def batch_feature(self, batch_ids):
        """Return a matrix (batch of feature vectors) corresponding to provided batch_ids
        """
        assert self.features is not None
        return self.features[batch_ids]



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import itertools
from collections import Counter


def ui_parser(tokens, line_idx, id_inline=False, **kwargs):
    if id_inline:
        return [(str(line_idx + 1), iid, 1.0) for iid in tokens]
    else:
        return [(tokens[0], iid, 1.0) for iid in tokens[1:]]


def uir_parser(tokens, **kwargs):
    return [(tokens[0], tokens[1], float(tokens[2]))]


def review_parser(tokens, **kwargs):
    return [(tokens[0], tokens[1], tokens[2])]


def uirt_parser(tokens, **kwargs):
    return [(tokens[0], tokens[1], float(tokens[2]), int(tokens[3]))]


def tup_parser(tokens, **kwargs):
    return [
        (
            tokens[0],
            tokens[1],
            [tuple(tup.split(kwargs.get("tup_sep"))) for tup in tokens[2:]],
        )
    ]


PARSERS = {"UI": ui_parser, "UIR": uir_parser, "UIRT": uirt_parser, "UITup": tup_parser, "UIReview": review_parser}


class Reader:
    """Reader class for reading data with different types of format.

    Parameters
    ----------
    user_set: set, default = None
        Set of users to be retained when reading data.
        If `None`, all users will be included.

    item_set: set, default = None
        Set of items to be retained when reading data.
        If `None`, all items will be included.

    min_user_freq: int, default = 1
        The minimum frequency of a user to be retained.
        If `min_user_freq = 1`, all users will be included.

    min_item_freq: int, default = 1
        The minimum frequency of an item to be retained.
        If `min_item_freq = 1`, all items will be included.

    bin_threshold: float, default = None
        The rating threshold to binarize rating values (turn explicit feedback to implicit feedback).
        For example, if `bin_threshold = 3.0`, all rating values >= 3.0 will be set to 1.0,
        and the rest (< 3.0) will be discarded.

    encoding: str, default = `utf-8`
        Encoding used to decode the file.

    errors: int, default = None
        Optional string that specifies how encoding errors are to be handled.
        Pass 'strict' to raise a ValueError exception if there is an encoding error
        (None has the same effect), or pass 'ignore' to ignore errors.
    """

    def __init__(
        self,
        user_set=None,
        item_set=None,
        min_user_freq=1,
        min_item_freq=1,
        bin_threshold=None,
        encoding="utf-8",
        errors=None,
    ):
        self.user_set = (
            user_set
            if (user_set is None or isinstance(user_set, set))
            else set(user_set)
        )
        self.item_set = (
            item_set
            if (item_set is None or isinstance(item_set, set))
            else set(item_set)
        )
        self.min_uf = min_user_freq
        self.min_if = min_item_freq
        self.bin_threshold = bin_threshold
        self.encoding = encoding
        self.errors = errors

    def _filter(self, tuples):
        if self.bin_threshold is not None:

            def binarize(t):
                t = list(t)
                t[2] = 1.0
                return tuple(t)

            tuples = [binarize(t) for t in tuples if t[2] >= self.bin_threshold]

        if self.user_set is not None:
            tuples = [t for t in tuples if t[0] in self.user_set]

        if self.item_set is not None:
            tuples = [t for t in tuples if t[1] in self.item_set]

        if self.min_uf > 1:
            user_freq = Counter(t[0] for t in tuples)
            tuples = [t for t in tuples if user_freq[t[0]] >= self.min_uf]

        if self.min_if > 1:
            item_freq = Counter(t[1] for t in tuples)
            tuples = [t for t in tuples if item_freq[t[1]] >= self.min_if]

        return tuples

    def read(
        self,
        fpath,
        fmt="UIR",
        sep="\t",
        skip_lines=0,
        id_inline=False,
        parser=None,
        **kwargs
    ):
        """Read data and parse line by line based on provided `fmt` or `parser`.

        Parameters
        ----------
        fpath: str
            Path to the data file.

        fmt: str, default: 'UIR'
            Line format to be parsed ('UIR' or 'UIRT').

        sep: str, default: '\t'
            The delimiter string.

        skip_lines: int, default: 0
            Number of first lines to skip

        id_inline: bool, default: False
            If `True`, user ids corresponding to the line numbers of the file,
            where all the ids in each line are item ids.

        parser: function, default: None
            Function takes a list of `str` tokenized by `sep` and
            returns a list of tuples which will be joined to the final results.
            If `None`, parser will be determined based on `fmt`.

        Returns
        -------
        tuples: list
            Data in the form of list of tuples. What inside each tuple
            depends on `parser` or `fmt`.

        """
        parser = PARSERS.get(fmt, None) if parser is None else parser
        if parser is None:
            raise ValueError(
                "Invalid line format: {}\n"
                "Only support: {}".format(fmt, PARSERS.keys())
            )

        with open(fpath, encoding=self.encoding, errors=self.errors) as f:
            tuples = [
                tup
                for idx, line in enumerate(itertools.islice(f, skip_lines, None))
                for tup in parser(
                    line.strip().split(sep), line_idx=idx, id_inline=id_inline, **kwargs
                )
            ]
            return self._filter(tuples)


def read_text(fpath, sep=None, encoding="utf-8", errors=None):
    """Read text file and return two lists of text documents and corresponding ids.
    If `sep` is None, only return one list containing elements are lines of text
    in the original file.

    Parameters
    ----------
    fpath: str
        Path to the data file

    sep: str, default = None
        The delimiter string used to split `id` and `text`. Each line is assumed
        containing an `id` followed by corresponding `text` document.
        If `None`, each line will be a `str` in returned list.

    encoding: str, default = `utf-8`
        Encoding used to decode the file.

    errors: int, default = None
        Optional string that specifies how encoding errors are to be handled.
        Pass 'strict' to raise a ValueError exception if there is an encoding error
        (None has the same effect), or pass 'ignore' to ignore errors.
    
    Returns
    -------
    texts, ids (optional): list, list
        Return list of text strings with corresponding indices (if `sep` is not None).
    """
    with open(fpath, encoding=encoding, errors=errors) as f:
        if sep is None:
            return [line.strip() for line in f]
        else:
            texts, ids = [], []
            for line in f:
                tokens = line.strip().split(sep)
                ids.append(tokens[0])
                texts.append(sep.join(tokens[1:]))
            return texts, ids



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Link to the data: https://www.kaggle.com/netflix-inc/netflix-prize-data/"""

from typing import List

from ..utils import validate_format
from ..utils import cache
from ..data import Reader

VALID_DATA_FORMATS = ['UIR', 'UIRT']
VARIANTS = ['original', 'small']
FNAME = {'small':'data_small', 'original':'data'}


def _load(fname, fmt='UIR', reader: Reader = None) -> List:
    """Load the Netflix dataset

    Parameters
    ----------
    fname: str, required
        Data file name.

    fmt: str, default: 'UIR'
        Data format to be returned.

    reader: `obj:cornac.data.Reader`, default: None
        Reader object used to read the data.

    Returns
    -------
    data: array-like
        Data in the form of a list of tuples depending on the given data format.

    """
    fmt = validate_format(fmt, VALID_DATA_FORMATS)
    fpath = cache(url='https://static.preferred.ai/cornac/datasets/netflix/{}.zip'.format(fname),
                  unzip=True, relative_path='netflix/{}.csv'.format(fname))
    reader = Reader() if reader is None else reader
    return reader.read(fpath, fmt, sep=',')


def load_feedback(fmt='UIR', variant='original', reader: Reader = None) -> List:
    """Load Netflix user-item ratings, scale: [1,5]

    Parameters
    ----------
    fmt: str, default: 'UIR'
        Data format to be returned.

    variant: str, optional, default: 'original'
        Specifies which Netflix dataset to load, one of ['original', 'small'].

    reader: `obj:cornac.data.Reader`, default: None
        Reader object used to read the data.

    Returns
    -------
    data: array-like
        Data in the form of a list of tuples depending on the given data format.

    """

    if variant not in VARIANTS:
        raise ValueError('variant must be one of {}.'.format(VARIANTS))

    return _load(FNAME[variant], fmt, reader)



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import itertools as it

import numpy as np
import torch
import torch.nn as nn
from tqdm.auto import trange


EPS = 1e-10

ACT = {
    "sigmoid": nn.Sigmoid(),
    "tanh": nn.Tanh(),
    "elu": nn.ELU(),
    "relu": nn.ReLU(),
    "relu6": nn.ReLU6(),
}


class BiVAE(nn.Module):
    def __init__(
        self,
        k,
        user_encoder_structure,
        item_encoder_structure,
        act_fn,
        likelihood,
        cap_priors,
        feature_dim,
        batch_size,
    ):
        super(BiVAE, self).__init__()

        self.mu_theta = torch.zeros((item_encoder_structure[0], k))  # n_users*k
        self.mu_beta = torch.zeros((user_encoder_structure[0], k))  # n_items*k

        self.theta = torch.randn(item_encoder_structure[0], k) * 0.01
        self.beta = torch.randn(user_encoder_structure[0], k) * 0.01
        torch.nn.init.kaiming_uniform_(self.theta, a=np.sqrt(5))

        self.likelihood = likelihood
        self.act_fn = ACT.get(act_fn, None)
        if self.act_fn is None:
            raise ValueError("Supported act_fn: {}".format(ACT.keys()))

        self.cap_priors = cap_priors
        if self.cap_priors.get("user", False):
            self.user_prior_encoder = nn.Linear(feature_dim.get("user"), k)
        if self.cap_priors.get("item", False):
            self.item_prior_encoder = nn.Linear(feature_dim.get("item"), k)

        # User Encoder
        self.user_encoder = nn.Sequential()
        for i in range(len(user_encoder_structure) - 1):
            self.user_encoder.add_module(
                "fc{}".format(i),
                nn.Linear(user_encoder_structure[i], user_encoder_structure[i + 1]),
            )
            self.user_encoder.add_module("act{}".format(i), self.act_fn)
        self.user_mu = nn.Linear(user_encoder_structure[-1], k)  # mu
        self.user_std = nn.Linear(user_encoder_structure[-1], k)

        # Item Encoder
        self.item_encoder = nn.Sequential()
        for i in range(len(item_encoder_structure) - 1):
            self.item_encoder.add_module(
                "fc{}".format(i),
                nn.Linear(item_encoder_structure[i], item_encoder_structure[i + 1]),
            )
            self.item_encoder.add_module("act{}".format(i), self.act_fn)
        self.item_mu = nn.Linear(item_encoder_structure[-1], k)  # mu
        self.item_std = nn.Linear(item_encoder_structure[-1], k)

    def to(self, device):
        self.beta = self.beta.to(device=device)
        self.theta = self.theta.to(device=device)
        self.mu_beta = self.mu_beta.to(device=device)
        self.mu_theta = self.mu_theta.to(device=device)
        return super(BiVAE, self).to(device)

    def encode_user_prior(self, x):
        h = self.user_prior_encoder(x)
        return h

    def encode_item_prior(self, x):
        h = self.item_prior_encoder(x)
        return h

    def encode_user(self, x):
        h = self.user_encoder(x)
        return self.user_mu(h), torch.sigmoid(self.user_std(h))

    def encode_item(self, x):
        h = self.item_encoder(x)
        return self.item_mu(h), torch.sigmoid(self.item_std(h))

    def decode_user(self, theta, beta):
        h = theta.mm(beta.t())
        return torch.sigmoid(h)

    def decode_item(self, theta, beta):
        h = beta.mm(theta.t())
        return torch.sigmoid(h)

    def reparameterize(self, mu, std):
        eps = torch.randn_like(mu)
        return mu + eps * std

    def forward(self, x, user=True, beta=None, theta=None):

        if user:
            mu, std = self.encode_user(x)
            theta = self.reparameterize(mu, std)
            return theta, self.decode_user(theta, beta), mu, std
        else:
            mu, std = self.encode_item(x)
            beta = self.reparameterize(mu, std)
            return beta, self.decode_item(theta, beta), mu, std

    def loss(self, x, x_, mu, mu_prior, std, kl_beta):
        # Likelihood
        ll_choices = {
            "bern": x * torch.log(x_ + EPS) + (1 - x) * torch.log(1 - x_ + EPS),
            "gaus": -(x - x_) ** 2,
            "pois": x * torch.log(x_ + EPS) - x_,
        }

        ll = ll_choices.get(self.likelihood, None)
        if ll is None:
            raise ValueError("Supported likelihoods: {}".format(ll_choices.keys()))

        ll = torch.sum(ll, dim=1)

        # KL term
        kld = -0.5 * (1 + 2.0 * torch.log(std) - (mu - mu_prior).pow(2) - std.pow(2))
        kld = torch.sum(kld, dim=1)

        return torch.mean(kl_beta * kld - ll)


def learn(
    bivae,
    train_set,
    n_epochs,
    batch_size,
    learn_rate,
    beta_kl,
    verbose,
    device=torch.device("cpu"),
    dtype=torch.float32,
):
    user_params = it.chain(
        bivae.user_encoder.parameters(),
        bivae.user_mu.parameters(),
        bivae.user_std.parameters(),
    )

    item_params = it.chain(
        bivae.item_encoder.parameters(),
        bivae.item_mu.parameters(),
        bivae.item_std.parameters(),
    )

    if bivae.cap_priors.get("user", False):
        user_params = it.chain(user_params, bivae.user_prior_encoder.parameters())
        user_features = train_set.user_feature.features[: train_set.num_users]

    if bivae.cap_priors.get("item", False):
        item_params = it.chain(item_params, bivae.item_prior_encoder.parameters())
        item_features = train_set.item_feature.features[: train_set.num_items]

    u_optimizer = torch.optim.Adam(params=user_params, lr=learn_rate)
    i_optimizer = torch.optim.Adam(params=item_params, lr=learn_rate)

    x = train_set.matrix.copy()
    x.data = np.ones_like(x.data)  # Binarize data
    tx = x.transpose()

    progress_bar = trange(1, n_epochs + 1, disable=not verbose)
    for _ in progress_bar:
        # item side
        i_sum_loss = 0.0
        i_count = 0
        for i_ids in train_set.item_iter(batch_size, shuffle=False):
            i_batch = tx[i_ids, :]
            i_batch = i_batch.A
            i_batch = torch.tensor(i_batch, dtype=dtype, device=device)

            # Reconstructed batch
            beta, i_batch_, i_mu, i_std = bivae(i_batch, user=False, theta=bivae.theta)

            i_mu_prior = 0.0  # zero mean for standard normal prior if not CAP prior
            if bivae.cap_priors.get("item", False):
                i_batch_f = item_features[i_ids]
                i_batch_f = torch.tensor(i_batch_f, dtype=dtype, device=device)
                i_mu_prior = bivae.encode_item_prior(i_batch_f)

            i_loss = bivae.loss(i_batch, i_batch_, i_mu, i_mu_prior, i_std, beta_kl)
            i_optimizer.zero_grad()
            i_loss.backward()
            i_optimizer.step()

            i_sum_loss += i_loss.data.item()
            i_count += len(i_batch)

            beta, _, i_mu, _ = bivae(i_batch, user=False, theta=bivae.theta)

            bivae.beta.data[i_ids] = beta.data
            bivae.mu_beta.data[i_ids] = i_mu.data

        # user side
        u_sum_loss = 0.0
        u_count = 0
        for u_ids in train_set.user_iter(batch_size, shuffle=False):
            u_batch = x[u_ids, :]
            u_batch = u_batch.A
            u_batch = torch.tensor(u_batch, dtype=dtype, device=device)

            # Reconstructed batch
            theta, u_batch_, u_mu, u_std = bivae(u_batch, user=True, beta=bivae.beta)

            u_mu_prior = 0.0  # zero mean for standard normal prior if not CAP prior
            if bivae.cap_priors.get("user", False):
                u_batch_f = user_features[u_ids]
                u_batch_f = torch.tensor(u_batch_f, dtype=dtype, device=device)
                u_mu_prior = bivae.encode_user_prior(u_batch_f)

            u_loss = bivae.loss(u_batch, u_batch_, u_mu, u_mu_prior, u_std, beta_kl)
            u_optimizer.zero_grad()
            u_loss.backward()
            u_optimizer.step()

            u_sum_loss += u_loss.data.item()
            u_count += len(u_batch)

            theta, _, u_mu, _ = bivae(u_batch, user=True, beta=bivae.beta)
            bivae.theta.data[u_ids] = theta.data
            bivae.mu_theta.data[u_ids] = u_mu.data

            progress_bar.set_postfix(
                loss_i=(i_sum_loss / i_count), loss_u=(u_sum_loss / (u_count))
            )

    # infer mu_beta
    for i_ids in train_set.item_iter(batch_size, shuffle=False):
        i_batch = tx[i_ids, :]
        i_batch = i_batch.A
        i_batch = torch.tensor(i_batch, dtype=dtype, device=device)

        beta, _, i_mu, _ = bivae(i_batch, user=False, theta=bivae.theta)
        bivae.mu_beta.data[i_ids] = i_mu.data

    # infer mu_theta
    for u_ids in train_set.user_iter(batch_size, shuffle=False):
        u_batch = x[u_ids, :]
        u_batch = u_batch.A
        u_batch = torch.tensor(u_batch, dtype=dtype, device=device)

        theta, _, u_mu, _ = bivae(u_batch, user=True, beta=bivae.beta)
        bivae.mu_theta.data[u_ids] = u_mu.data

    return bivae



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Link to the data: https://grouplens.org/datasets/movielens/"""

from typing import List
from collections import namedtuple

from ..utils import validate_format
from ..utils import cache
from ..data import Reader
from ..data.reader import read_text


VALID_DATA_FORMATS = ["UIR", "UIRT"]

MovieLens = namedtuple("MovieLens", ["url", "unzip", "path", "sep", "skip"])
ML_DATASETS = {
    "100K": MovieLens(
        "http://files.grouplens.org/datasets/movielens/ml-100k/u.data",
        False,
        "ml-100k/u.data",
        "\t",
        0,
    ),
    "1M": MovieLens(
        "http://files.grouplens.org/datasets/movielens/ml-1m.zip",
        True,
        "ml-1m/ratings.dat",
        "::",
        0,
    ),
    "10M": MovieLens(
        "http://files.grouplens.org/datasets/movielens/ml-10m.zip",
        True,
        "ml-10M100K/ratings.dat",
        "::",
        0,
    ),
    "20M": MovieLens(
        "http://files.grouplens.org/datasets/movielens/ml-20m.zip",
        True,
        "ml-20m/ratings.csv",
        ",",
        1,
    ),
}


def load_feedback(fmt="UIR", variant="100K", reader=None):
    """Load the user-item ratings of one of the MovieLens datasets

    Parameters
    ----------
    fmt: str, default: 'UIR'
        Data format to be returned, one of ['UIR', 'UIRT'].

    variant: str, optional, default: '100K'
        Specifies which MovieLens dataset to load, one of ['100K', '1M', '10M', '20M'].

    reader: `obj:cornac.data.Reader`, optional, default: None
        Reader object used to read the data.

    Returns
    -------
    data: array-like
        Data in the form of a list of tuples depending on the given data format.
    """

    fmt = validate_format(fmt, VALID_DATA_FORMATS)

    ml = ML_DATASETS.get(variant.upper(), None)
    if ml is None:
        raise ValueError("variant must be one of {}.".format(ML_DATASETS.keys()))

    fpath = cache(url=ml.url, unzip=ml.unzip, relative_path=ml.path)
    reader = Reader() if reader is None else reader
    return reader.read(fpath, fmt, sep=ml.sep, skip_lines=ml.skip)


def load_plot():
    """Load the plots of movies provided @ http://dm.postech.ac.kr/~cartopy/ConvMF/

    Returns
    -------
    texts: List
        List of text documents, one per item.

    ids: List
        List of item ids aligned with indices in `texts`.
    """
    fpath = cache(
        url="https://static.preferred.ai/cornac/datasets/movielens/ml_plot.zip",
        unzip=True,
        relative_path="movielens/ml_plot.dat",
    )
    texts, ids = read_text(fpath, sep="::")
    return texts, ids



import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers, initializers
from tensorflow.python.keras.preprocessing.sequence import pad_sequences

from ...utils import get_rng
from ...utils.init_utils import uniform
from ..narre.narre import TextProcessor, AddGlobalBias


def get_data(batch_ids, train_set, max_text_length, by="user", max_num_review=32):
    batch_reviews, batch_num_reviews = [], []
    review_group = (
        train_set.review_text.user_review
        if by == "user"
        else train_set.review_text.item_review
    )
    for idx in batch_ids:
        review_ids = []
        for inc, (jdx, review_idx) in enumerate(review_group[idx].items()):
            if max_num_review is not None and inc == max_num_review:
                break
            review_ids.append(review_idx)
        reviews = train_set.review_text.batch_seq(
            review_ids, max_length=max_text_length
        )
        batch_reviews.append(reviews)
        batch_num_reviews.append(len(reviews))
    batch_reviews = pad_sequences(batch_reviews, maxlen=max_num_review, padding="post")
    batch_num_reviews = np.array(batch_num_reviews).astype(np.int32)
    batch_ratings = (
        np.zeros((len(batch_ids), train_set.num_items), dtype=np.float32)
        if by == "user"
        else np.zeros((len(batch_ids), train_set.num_users), dtype=np.float32)
    )
    rating_group = train_set.user_data if by == "user" else train_set.item_data
    for batch_inc, idx in enumerate(batch_ids):
        jds, ratings = rating_group[idx]
        for jdx, rating in zip(jds, ratings):
            batch_ratings[batch_inc, jdx] = rating
    return batch_reviews, batch_num_reviews, batch_ratings

class Model(keras.Model):
    def __init__(self, n_users, n_items, n_vocab, global_mean, embedding_matrix,
                 n_factors=32, embedding_size=100, id_embedding_size=32,
                 attention_size=16, kernel_sizes=[3], n_filters=64,
                 n_user_mlp_factors=128, n_item_mlp_factors=128,
                 dropout_rate=0.5, max_text_length=50):
        super().__init__()
        self.l_user_review_embedding = layers.Embedding(n_vocab, embedding_size, embeddings_initializer=embedding_matrix, mask_zero=True, name="user_review_embedding")
        self.l_item_review_embedding = layers.Embedding(n_vocab, embedding_size, embeddings_initializer=embedding_matrix, mask_zero=True, name="item_review_embedding")
        self.l_user_embedding = layers.Embedding(n_users, id_embedding_size, embeddings_initializer="uniform", name="user_embedding")
        self.l_item_embedding = layers.Embedding(n_items, id_embedding_size, embeddings_initializer="uniform", name="item_embedding")
        self.user_bias = layers.Embedding(n_users, 1, embeddings_initializer=tf.initializers.Constant(0.1), name="user_bias")
        self.item_bias = layers.Embedding(n_items, 1, embeddings_initializer=tf.initializers.Constant(0.1), name="item_bias")
        self.user_text_processor = TextProcessor(max_text_length, filters=n_filters, kernel_sizes=kernel_sizes, dropout_rate=dropout_rate, name='user_text_processor')
        self.item_text_processor = TextProcessor(max_text_length, filters=n_filters, kernel_sizes=kernel_sizes, dropout_rate=dropout_rate, name='item_text_processor')

        self.l_user_mlp = keras.models.Sequential([
            layers.Dense(n_user_mlp_factors, input_dim=n_items, activation="relu"),
            layers.Dense(n_user_mlp_factors // 2, activation="relu"),
            layers.Dense(n_filters, activation="relu"),
            layers.BatchNormalization(),
        ])
        self.l_item_mlp = keras.models.Sequential([
            layers.Dense(n_item_mlp_factors, input_dim=n_users, activation="relu"),
            layers.Dense(n_item_mlp_factors // 2, activation="relu"),
            layers.Dense(n_filters, activation="relu"),
            layers.BatchNormalization(),
        ])
        self.a_user = keras.models.Sequential([
            layers.Dense(attention_size, activation="relu", use_bias=True),
            layers.Dense(1, activation=None, use_bias=True)
        ])
        self.user_attention = layers.Softmax(axis=1, name="user_attention")
        self.a_item = keras.models.Sequential([
            layers.Dense(attention_size, activation="relu", use_bias=True),
            layers.Dense(1, activation=None, use_bias=True)
        ])
        self.item_attention = layers.Softmax(axis=1, name="item_attention")
        self.ou_dropout = layers.Dropout(rate=dropout_rate)
        self.oi_dropout = layers.Dropout(rate=dropout_rate)
        self.ou = layers.Dense(n_factors, use_bias=True, name="ou")
        self.oi = layers.Dense(n_factors, use_bias=True, name="oi")
        self.W1 = layers.Dense(1, activation=None, use_bias=False, name="W1")
        self.add_global_bias = AddGlobalBias(init_value=global_mean, name="global_bias")

    def call(self, inputs, training=False):
        i_user_id, i_item_id, i_user_rating, i_user_review, i_user_num_reviews, i_item_rating, i_item_review, i_item_num_reviews = inputs
        user_review_h = self.user_text_processor(self.l_user_review_embedding(i_user_review), training=training)
        item_review_h = self.item_text_processor(self.l_item_review_embedding(i_item_review), training=training)
        user_rating_h = self.l_user_mlp(i_user_rating)
        item_rating_h = self.l_item_mlp(i_item_rating)
        a_user = self.a_user(
            tf.multiply(
                user_review_h,
                tf.expand_dims(user_rating_h, 1)
            )
        )
        a_user_masking = tf.expand_dims(tf.sequence_mask(tf.reshape(i_user_num_reviews, [-1]), maxlen=i_user_review.shape[1]), -1)
        user_attention = self.user_attention(a_user, a_user_masking)
        a_item = self.a_item(
            tf.multiply(
                item_review_h,
                tf.expand_dims(item_rating_h, 1)
            )
        )
        a_item_masking = tf.expand_dims(tf.sequence_mask(tf.reshape(i_item_num_reviews, [-1]), maxlen=i_item_review.shape[1]), -1)
        item_attention = self.item_attention(a_item, a_item_masking)
        ou = tf.multiply(user_attention, user_review_h)
        ou = tf.reduce_sum(ou, 1)
        if training:
            ou = self.ou_dropout(ou, training=training)
        ou = self.ou(ou)
        oi = tf.multiply(item_attention, item_review_h)
        oi = tf.reduce_sum(oi, 1)
        if training:
            oi = self.oi_dropout(oi, training=training)
        oi = self.oi(oi)
        pu = tf.concat([
            user_rating_h,
            ou,
            self.l_user_embedding(i_user_id)
        ], axis=-1)
        qi = tf.concat([
            item_rating_h,
            oi,
            self.l_item_embedding(i_item_id)
        ], axis=-1)
        h0 = tf.multiply(pu, qi)
        r = self.add_global_bias(
            tf.add_n([
                self.W1(h0),
                self.user_bias(i_user_id),
                self.item_bias(i_item_id)
            ])
        )
        return r

class HRDRModel:
    def __init__(self, n_users, n_items, vocab, global_mean,
                 n_factors=32, embedding_size=100, id_embedding_size=32,
                 attention_size=16, kernel_sizes=[3], n_filters=64,
                 n_user_mlp_factors=128, n_item_mlp_factors=128,
                 dropout_rate=0.5, max_text_length=50, max_num_review=32,
                 pretrained_word_embeddings=None, verbose=False, seed=None):
        self.n_users = n_users
        self.n_items = n_items
        self.n_vocab = vocab.size
        self.global_mean = global_mean
        self.n_factors = n_factors
        self.embedding_size = embedding_size
        self.id_embedding_size = id_embedding_size
        self.attention_size = attention_size
        self.kernel_sizes = kernel_sizes
        self.n_filters = n_filters
        self.n_user_mlp_factors = n_user_mlp_factors
        self.n_item_mlp_factors = n_item_mlp_factors
        self.dropout_rate = dropout_rate
        self.max_text_length = max_text_length
        self.max_num_review = max_num_review
        self.verbose = verbose
        if seed is not None:
            self.rng = get_rng(seed)
            tf.random.set_seed(seed)

        embedding_matrix = uniform(shape=(self.n_vocab, self.embedding_size), low=-0.5, high=0.5, random_state=self.rng)
        embedding_matrix[:4, :] = np.zeros((4, self.embedding_size))
        if pretrained_word_embeddings is not None:
            oov_count = 0
            for word, idx in vocab.tok2idx.items():
                embedding_vector = pretrained_word_embeddings.get(word)
                if embedding_vector is not None:
                    embedding_matrix[idx] = embedding_vector
                else:
                    oov_count += 1
            if self.verbose:
                print("Number of OOV words: %d" % oov_count)

        embedding_matrix = initializers.Constant(embedding_matrix)
        self.graph = Model(
            self.n_users, self.n_items, self.n_vocab, self.global_mean, embedding_matrix,
            self.n_factors, self.embedding_size, self.id_embedding_size,
            self.attention_size, self.kernel_sizes, self.n_filters,
            self.n_user_mlp_factors, self.n_item_mlp_factors,
            self.dropout_rate, self.max_text_length
        )

    def get_weights(self, train_set, batch_size=64):
        P = np.zeros((self.n_users, self.n_filters + self.n_factors + self.id_embedding_size))
        Q = np.zeros((self.n_items, self.n_filters + self.n_factors + self.id_embedding_size))
        A = np.zeros((self.n_items, self.max_num_review))
        for batch_users in train_set.user_iter(batch_size, shuffle=False):
            i_user_review, i_user_num_reviews, i_user_rating = get_data(batch_users, train_set, self.max_text_length, by='user', max_num_review=self.max_num_review)
            user_review_embedding = self.graph.l_user_review_embedding(i_user_review)
            user_review_h = self.graph.user_text_processor(user_review_embedding, training=False)
            user_rating_h = self.graph.l_user_mlp(i_user_rating)
            a_user = self.graph.a_user(
                tf.multiply(
                    user_review_h,
                    tf.expand_dims(user_rating_h, 1)
                )
            )
            a_user_masking = tf.expand_dims(tf.sequence_mask(tf.reshape(i_user_num_reviews, [-1]), maxlen=i_user_review.shape[1]), -1)
            user_attention = self.graph.user_attention(a_user, a_user_masking)
            ou = self.graph.ou(tf.reduce_sum(tf.multiply(user_attention, user_review_h), 1))
            pu = tf.concat([
                user_rating_h,
                ou,
                self.graph.l_user_embedding(batch_users)
            ], axis=-1)
            P[batch_users] = pu.numpy()
        for batch_items in train_set.item_iter(batch_size, shuffle=False):
            i_item_review, i_item_num_reviews, i_item_rating = get_data(batch_items, train_set, self.max_text_length, by='item', max_num_review=self.max_num_review)
            item_review_embedding = self.graph.l_item_review_embedding(i_item_review)
            item_review_h = self.graph.item_text_processor(item_review_embedding, training=False)
            item_rating_h = self.graph.l_item_mlp(i_item_rating)
            a_item = self.graph.a_item(
                tf.multiply(
                    item_review_h,
                    tf.expand_dims(item_rating_h, 1)
                )
            )
            a_item_masking = tf.expand_dims(tf.sequence_mask(tf.reshape(i_item_num_reviews, [-1]), maxlen=i_item_review.shape[1]), -1)
            item_attention = self.graph.item_attention(a_item, a_item_masking)
            oi = self.graph.oi(tf.reduce_sum(tf.multiply(item_attention, item_review_h), 1))
            qi = tf.concat([
                item_rating_h,
                oi,
                self.graph.l_item_embedding(batch_items)
            ], axis=-1)
            Q[batch_items] = qi.numpy()
            A[batch_items, :item_attention.shape[1]] = item_attention.numpy().reshape(item_attention.shape[:2])
        W1 = self.graph.W1.get_weights()[0]
        bu = self.graph.user_bias.get_weights()[0]
        bi = self.graph.item_bias.get_weights()[0]
        mu = self.graph.add_global_bias.get_weights()[0][0]
        return P, Q, W1, bu, bi, mu, A



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import os
import shutil
import zipfile
import tarfile
from urllib import request

from tqdm.auto import tqdm


def _urlretrieve(url, fpath):
    """Retrieve data from given url

    Parameters
    ----------
    url: str
        The url to the data.

    fpath: str
        The path to file where data is stored.

    """
    opener = request.build_opener()
    opener.addheaders = [("User-agent", "Mozilla/5.0")]

    with tqdm(unit="B", unit_scale=True) as progress:

        def report(chunk, chunksize, total):
            progress.total = total
            progress.update(chunksize)

        request.install_opener(opener)
        request.urlretrieve(url, fpath, reporthook=report)


def _extract_archive(file_path, extract_path="."):
    """Extracts an archive.
    """
    for archive_type in ["zip", "tar"]:
        if archive_type == "zip":
            open_fn = zipfile.ZipFile
            is_match_fn = zipfile.is_zipfile
        elif archive_type == "tar":
            open_fn = tarfile.open
            is_match_fn = tarfile.is_tarfile

        if is_match_fn(file_path):
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(extract_path)
                except (tarfile.TarError, RuntimeError, KeyboardInterrupt):
                    if os.path.exists(extract_path):
                        if os.path.isfile(extract_path):
                            os.remove(extract_path)
                        else:
                            shutil.rmtree(extract_path)
                    raise


def get_cache_path(relative_path, cache_dir=None):
    """Return the absolute path to the cached data file
    """
    if cache_dir is None and os.access(os.path.expanduser("~"), os.W_OK):
        cache_dir = os.path.join(os.path.expanduser("~"), ".cornac")
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

    if not os.access(cache_dir, os.W_OK):
        cache_dir = os.path.join("/tmp", ".cornac")
    cache_path = os.path.join(cache_dir, relative_path)

    if not os.path.exists(os.path.dirname(cache_path)):
        os.makedirs(os.path.dirname(cache_path))

    return cache_path, cache_dir


def cache(url, unzip=False, relative_path=None, cache_dir=None):
    """Download the data and cache to file

    Parameters
    ----------
    url: str
        The url to the data.

    unzip: bool, optional, default: False
        Whether the data is a zip file and going to be unzipped after the download.

    relative_path: str
        Relative path to the data file after finishing the download.
        If unzip=True, relative_path is the path to unzipped file.

    cache_dir: str, optional, default: None
        The path to cache folder. If `None`, either ~/.cornac or /tmp/.cornac will be used.

    """
    if relative_path is None:
        relative_path = url.split("/")[-1]
    cache_path, cache_dir = get_cache_path(relative_path, cache_dir)
    if os.path.exists(cache_path):
        return cache_path

    print("Data from", url)
    print("will be cached into", cache_path)

    if unzip:
        tmp_path = os.path.join(cache_dir, "file.tmp")
        _urlretrieve(url, tmp_path)
        print("Unzipping ...")
        _extract_archive(tmp_path, cache_dir)
        os.remove(tmp_path)
    else:
        _urlretrieve(url, cache_path)

    print("File cached!")
    return cache_path



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from tqdm.auto import trange

from ..recommender import Recommender
from ...exception import ScoreException
from ...utils import get_rng
from ...utils.init_utils import xavier_uniform


class CTR(Recommender):
    """Collaborative Topic Regression.

    Parameters
    ----------
    name: string, default: 'CTR'
        The name of the recommender model.

    k: int, optional, default: 200
        The dimension of the latent factors.

    max_iter: int, optional, default: 100
        Maximum number of iterations or the number of epochs for SGD.

    lambda_u: float, optional, default: 0.01
        The regularization parameter for users.

    lambda_v: float, optional, default: 0.01
        The regularization parameter for items.

    a: float, optional, default: 1
        The confidence of observed ratings.

    b: float, optional, default: 0.01
        The confidence of unseen ratings.

    eta: float, optional, default: 0.01
        Added value for smoothing phi.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model already 
        pre-trained (U and V are not None).

    init_params: dictionary, optional, default: None
        List of initial parameters, e.g., init_params = {'U':U, 'V':V}

        U: ndarray, shape (n_users,k)
            The user latent factors, optional initialization via init_params.
        V: ndarray, shape (n_items,k)
            The item latent factors, optional initialization via init_params.

    seed: int, optional, default: None
        Random seed for weight initialization.

    References
    ----------
    Wang, Chong, and David M. Blei. "Collaborative topic modeling for recommending scientific articles."
    Proceedings of the 17th ACM SIGKDD international conference on Knowledge discovery and data mining. ACM, 2011.

    """

    def __init__(
        self,
        name="CTR",
        k=200,
        lambda_u=0.01,
        lambda_v=0.01,
        eta=0.01,
        a=1,
        b=0.01,
        max_iter=100,
        trainable=True,
        verbose=True,
        init_params=None,
        seed=None,
    ):
        super().__init__(name=name, trainable=trainable, verbose=verbose)
        self.k = k
        self.lambda_u = lambda_u
        self.lambda_v = lambda_v
        self.a = a
        self.b = b
        self.eta = eta
        self.name = name
        self.max_iter = max_iter
        self.verbose = verbose
        self.seed = seed

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.U = self.init_params.get("U", None)
        self.V = self.init_params.get("V", None)

    def _init(self):
        rng = get_rng(self.seed)
        self.n_item = self.train_set.num_items
        self.n_user = self.train_set.num_users

        if self.U is None:
            self.U = xavier_uniform((self.n_user, self.k), rng)
        if self.V is None:
            self.V = xavier_uniform((self.n_item, self.k), rng)

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        self._init()

        if self.trainable:
            self._fit_ctr()

        return self

    @staticmethod
    def _build_data(csr_mat):
        index_list = []
        rating_list = []
        for i in range(csr_mat.shape[0]):
            j, k = csr_mat.indptr[i], csr_mat.indptr[i + 1]
            index_list.append(csr_mat.indices[j:k])
            rating_list.append(csr_mat.data[j:k])
        return index_list, rating_list

    def _fit_ctr(self,):
        from .ctr import Model

        user_data = self._build_data(self.train_set.matrix)
        item_data = self._build_data(self.train_set.matrix.T.tocsr())

        bow_mat = self.train_set.item_text.batch_bow(
            np.arange(self.n_item), keep_sparse=True
        )
        doc_ids, doc_cnt = self._build_data(bow_mat)  # bag of word feature

        self.model = Model(
            n_user=self.n_user,
            n_item=self.n_item,
            U=self.U,
            V=self.V,
            k=self.k,
            n_vocab=self.train_set.item_text.vocab.size,
            lambda_u=self.lambda_u,
            lambda_v=self.lambda_v,
            a=self.a,
            b=self.b,
            max_iter=self.max_iter,
            seed=self.seed,
        )

        loop = trange(self.max_iter, disable=not self.verbose)
        for _ in loop:
            cf_loss = self.model.update_cf(
                user_data=user_data, item_data=item_data
            )  # u and v updating
            lda_loss = self.model.update_theta(doc_ids=doc_ids, doc_cnt=doc_cnt)
            self.model.update_beta()
            loop.set_postfix(cf_loss=cf_loss, lda_likelihood=-lda_loss)

        if self.verbose:
            print("Learning completed!")

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items
        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            known_item_scores = self.V.dot(self.U[user_idx, :])
            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )
            user_pred = self.V[item_idx, :].dot(self.U[user_idx, :])
            return user_pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import os
from datetime import datetime

from .result import ExperimentResult
from .result import CVExperimentResult
from ..metrics.rating import RatingMetric
from ..metrics.ranking import RankingMetric
from ..models.recommender import Recommender


class Experiment:
    """ Experiment Class

    Parameters
    ----------
    eval_method: :obj:`<cornac.eval_methods.BaseMethod>`, required
        The evaluation method (e.g., RatioSplit).

    models: array of :obj:`<cornac.models.Recommender>`, required
        A collection of recommender models to evaluate, e.g., [C2PF, HPF, PMF].

    metrics: array of :obj:{`<cornac.metrics.RatingMetric>`, `<cornac.metrics.RankingMetric>`}, required
        A collection of metrics to use to evaluate the recommender models, \
        e.g., [NDCG, MRR, Recall].

    user_based: bool, optional, default: True
        This parameter is only useful if you are considering rating metrics. When True, first the average performance \
        for every user is computed, then the obtained values are averaged to return the final result.
        If `False`, results will be averaged over the number of ratings.

    show_validation: bool, optional, default: True 
        Whether to show the results on validation set (if exists).

    save_dir: str, optional, default: None
        Path to a directory for storing trained models and logs. If None, 
        models will NOT be stored and logs will be saved in the current working directory.

    Attributes
    ----------
    result: array of :obj:`<cornac.experiment.result.Result>`, default: None
        This attribute contains the results per-model of your experiment 
        on the test set, initially it is set to None.
    
    val_result: array of :obj:`<cornac.experiment.result.Result>`, default: None
        This attribute contains the results per-model of your experiment
        on the validation set (if exists), initially it is set to None.

    """

    def __init__(
        self,
        eval_method,
        models,
        metrics,
        user_based=True,
        show_validation=True,
        verbose=False,
        save_dir=None,
    ):
        self.eval_method = eval_method
        self.models = self._validate_models(models)
        self.metrics = self._validate_metrics(metrics)
        self.user_based = user_based
        self.show_validation = show_validation
        self.verbose = verbose
        self.save_dir = save_dir
        self.result = None
        self.val_result = None

    @staticmethod
    def _validate_models(input_models):
        if not hasattr(input_models, "__len__"):
            raise ValueError(
                "models have to be an array but {}".format(type(input_models))
            )

        valid_models = []
        for model in input_models:
            if isinstance(model, Recommender):
                valid_models.append(model)
        return valid_models

    @staticmethod
    def _validate_metrics(input_metrics):
        if not hasattr(input_metrics, "__len__"):
            raise ValueError(
                "metrics have to be an array but {}".format(type(input_metrics))
            )

        valid_metrics = []
        for metric in input_metrics:
            if isinstance(metric, RatingMetric) or isinstance(metric, RankingMetric):
                valid_metrics.append(metric)
        return valid_metrics

    def _create_result(self):
        from ..eval_methods.cross_validation import CrossValidation
        from ..eval_methods.propensity_stratified_evaluation import (
            PropensityStratifiedEvaluation,
        )

        if isinstance(self.eval_method, CrossValidation) or isinstance(
            self.eval_method, PropensityStratifiedEvaluation
        ):
            self.result = CVExperimentResult()
        else:
            self.result = ExperimentResult()
            if self.show_validation and self.eval_method.val_set is not None:
                self.val_result = ExperimentResult()

    def run(self):
        """Run the Cornac experiment"""
        self._create_result()

        for model in self.models:
            test_result, val_result = self.eval_method.evaluate(
                model=model,
                metrics=self.metrics,
                user_based=self.user_based,
                show_validation=self.show_validation,
            )

            self.result.append(test_result)
            if self.val_result is not None:
                self.val_result.append(val_result)

            if not isinstance(self.result, CVExperimentResult):
                model.save(self.save_dir)

        output = ""
        if self.val_result is not None:
            output += "\nVALIDATION:\n...\n{}".format(self.val_result)
        output += "\nTEST:\n...\n{}".format(self.result)

        print(output)

        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S-%f")
        save_dir = "." if self.save_dir is None else self.save_dir
        output_file = os.path.join(save_dir, "CornacExp-{}.log".format(timestamp))
        with open(output_file, "w") as f:
            f.write(output)


# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================


import numpy as np
from itertools import product

from .models import Recommender
from .metrics import RatingMetric, RankingMetric
from .eval_methods import rating_eval, ranking_eval
from .utils import get_rng


__all__ = ["Discrete", "Continuous", "GridSearch", "RandomSearch"]


class SearchDomain(object):
    """Domain of a parameter to search on.
    
    Parameters
    ----------------
    name: str, required
        Name of the parameter.

    """

    def __init__(self, name):
        self.name = name

    def _sample(self, rng):
        """Sample a value of parameter used for RandomSearch"""
        raise NotImplementedError()


class Discrete(SearchDomain):
    """Domain of a parameter with a set of discrete values.
    
    Parameters
    ----------------
    name: str, required
        Name of the parameter.
        
    values: list, required
        List of values to be searched.

    """

    def __init__(self, name, values):
        super().__init__(name=name)
        self.values = values

    def _sample(self, rng):
        """Sample a value of parameter used for RandomSearch"""
        return rng.choice(self.values)


class Continuous(SearchDomain):
    """Domain of a parameter with continuous values within a range of [low, high).
    
    Parameters
    ----------------
    name: str, required
        Name of the parameter.
        
    low: float, default: 0.0
        Lower bound of the searched values (included).

    high: float, default: 1.0
        Upper bound of the searched values (excluded).
            
    """

    def __init__(self, name, low=0.0, high=1.0):
        super().__init__(name=name)
        self.low = low
        self.high = high

    def _sample(self, rng):
        """Sample a value of parameter used for RandomSearch"""
        return rng.uniform(low=self.low, high=self.high)


class BaseSearch(Recommender):
    """Base class for doing parameter search.
    
    Parameters
    ----------------
    model: :obj:`cornac.models.Recommender`, required
        Base recommender model to be tuned.

    space: list, required
        Parameter space to be searched on.
        It's a list of :obj:`cornac.hyperopt.SearchDomain`.
    
    metric: :obj:`cornac.metrics.RatingMetric` or :obj:`cornac.metrics.RankingMetric`, required
        Scoring metric to measure the performance and rank the parameter settings.

    eval_method: :obj:`cornac.eval_methods.BaseMethod`, required
        Evaluation method is being used. 
        
    name: str, default: 'BaseSearch'
        The name of the searching strategy.
        
    """

    def __init__(self, model, space, metric, eval_method, name="BaseSearch"):
        super().__init__(name=name, verbose=model.verbose)
        self.model = model
        self.space = sorted(space, key=lambda x: x.name)  # for reproducibility
        self.metric = metric
        self.eval_method = eval_method

    def _build_param_set(self):
        """Generate searching points"""
        raise NotImplementedError()

    def fit(self, train_set, val_set=None):
        """Doing hyper-parameter search"""
        assert val_set is not None
        Recommender.fit(self, train_set, val_set)

        param_set = self._build_param_set()
        compare_op = np.greater if self.metric.higher_better else np.less
        self.best_score = -np.inf if self.metric.higher_better else np.inf
        self.best_model = None
        self.best_params = None

        # this can be parallelized if needed
        # keep it simple because multimodal algorithms are usually resource-hungry
        for params in param_set:
            if self.verbose:
                print("Evaluating: {}".format(params))

            model = self.model.clone(params).fit(train_set, val_set)

            if isinstance(self.metric, RatingMetric):
                score = rating_eval(model, [self.metric], val_set)[0][0]
            else:
                score = ranking_eval(
                    model,
                    [self.metric],
                    train_set,
                    val_set,
                    rating_threshold=self.eval_method.rating_threshold,
                    exclude_unknowns=self.eval_method.exclude_unknowns,
                    verbose=False,
                )[0][0]

            if compare_op(score, self.best_score):
                self.best_score = score
                self.best_model = model
                self.best_params = params

            del model

        if self.verbose:
            print("Best parameter settings: {}".format(self.best_params))
            print("{} = {:.4f}".format(self.metric.name, self.best_score))

        return self

    def score(self, user_idx, item_idx=None):
        """Scoring using the best searched model"""
        return self.best_model.score(user_idx, item_idx)


class GridSearch(BaseSearch):
    """Parameter searching on a grid.
    
    Parameters
    ----------------
    model: :obj:`cornac.models.Recommender`, required
        Base recommender model to be tuned.

    space: list, required
        Parameter space to be searched on.
        It's a list of :obj:`cornac.hyperopt.SearchDomain`.
    
    metric: :obj:`cornac.metrics.RatingMetric` or :obj:`cornac.metrics.RankingMetric`, required
        Scoring metric to measure the performance and rank the parameter settings.

    eval_method: :obj:`cornac.eval_methods.BaseMethod`, required
        Evaluation method is being used. 
        
    """

    def __init__(self, model, space, metric, eval_method):
        super().__init__(
            model,
            self._validate(space),
            metric,
            eval_method,
            name="GridSearch_{}".format(model.name),
        )

    @staticmethod
    def _validate(space):
        """GridSearch only accepts Discrete search domain"""
        for domain in space:
            if isinstance(domain, Discrete):
                continue

            raise ValueError(
                "GridSearch only supports Discrete domain but {} is not!\n\
                    Please consider using RandomSearch instead.".format(
                    domain.name
                )
            )

        return space

    def _build_param_set(self):
        """Generate searching points"""
        param_set = []
        keys = [d.name for d in self.space]
        for params in product(*[sorted(d.values) for d in self.space]):
            param_set.append(dict(zip(keys, params)))
        return param_set


class RandomSearch(BaseSearch):
    """Parameter searching with random strategy.
    
    Parameters
    ----------------
    model: :obj:`cornac.models.Recommender`, required
        Base recommender model to be tuned.

    space: list, required
        Parameter space to be searched on.
        It's a list of :obj:`cornac.hyperopt.SearchDomain`.
    
    metric: :obj:`cornac.metrics.RatingMetric` or :obj:`cornac.metrics.RankingMetric`, required
        Scoring metric to measure the performance and rank the parameter settings.

    eval_method: :obj:`cornac.eval_methods.BaseMethod`, required
        Evaluation method is being used. 

    n_trails: int, default: 10
        Number of trails for random search.

    """

    def __init__(self, model, space, metric, eval_method, n_trails=10):
        super().__init__(
            model, space, metric, eval_method, name="RandomSearch_{}".format(model.name)
        )
        self.n_trails = n_trails

    def _build_param_set(self):
        """Generate searching points"""
        param_set = []
        keys = [d.name for d in self.space]
        rng = get_rng(self.model.seed)
        while len(param_set) < self.n_trails:
            params = [d._sample(rng) for d in self.space]
            param_set.append(dict(zip(keys, params)))
        return param_set



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from tqdm.auto import trange

from ...utils import estimate_batches

torch.set_default_dtype(torch.float32)

EPS = 1e-10

ACT = {
    "sigmoid": nn.Sigmoid(),
    "tanh": nn.Tanh(),
    "elu": nn.ELU(),
    "relu": nn.ReLU(),
    "relu6": nn.ReLU6(),
}


class CVAE(nn.Module):
    def __init__(self, z_dim, h_dim, ae_structure_z, ae_structure_h, act_fn, likelihood):
        super(CVAE, self).__init__()

        self.likelihood = likelihood
        self.act_fn = ACT.get(act_fn, None)
        if self.act_fn is None:
            raise ValueError("Supported act_fn: {}".format(ACT.keys()))

        # Encoder for preference data y, q(z|y)
        self.encoder_qz = nn.Sequential()
        for i in range(len(ae_structure_z) - 1):
            self.encoder_qz.add_module(
                "fc{}".format(i), nn.Linear(ae_structure_z[i], ae_structure_z[i + 1])
            )
            self.encoder_qz.add_module("act{}".format(i), self.act_fn)
        self.enc_qz_mu = nn.Linear(ae_structure_z[-1], z_dim)  # mu
        self.enc_qz_logvar = nn.Linear(ae_structure_z[-1], z_dim)  # logvar

        # Encoder for auxiliary data x, q(h|x)
        self.encoder_qhx = nn.Sequential()
        for i in range(len(ae_structure_h) - 1):
            self.encoder_qhx.add_module(
                "fc{}".format(i), nn.Linear(ae_structure_h[i], ae_structure_h[i + 1])
            )
            self.encoder_qhx.add_module("act{}".format(i), self.act_fn)
        self.enc_qhx_mu = nn.Linear(ae_structure_h[-1], h_dim)  # mu
        self.enc_qhx_logvar = nn.Linear(ae_structure_h[-1], h_dim)  # logvar

        # Encoder for preference data y, q(h|y)
        self.encoder_qhy = nn.Sequential()
        for i in range(len(ae_structure_z) - 1):
            self.encoder_qhy.add_module(
                "fc{}".format(i), nn.Linear(ae_structure_z[i], ae_structure_z[i + 1])
            )
            self.encoder_qhy.add_module("act{}".format(i), self.act_fn)
        self.enc_qhy_mu = nn.Linear(ae_structure_z[-1], h_dim)  # mu
        self.enc_qhy_logvar = nn.Linear(ae_structure_z[-1], h_dim)  # logvar

        # Encoder for auxiliary data x, p(h|x)
        self.encoder_phx = nn.Sequential()
        for i in range(len(ae_structure_h) - 1):
            self.encoder_phx.add_module(
                "fc{}".format(i), nn.Linear(ae_structure_h[i], ae_structure_h[i + 1])
            )
            self.encoder_phx.add_module("act{}".format(i), self.act_fn)
        self.enc_ph_mu = nn.Linear(ae_structure_h[-1], h_dim)  # mu
        self.enc_ph_logvar = nn.Linear(ae_structure_h[-1], h_dim)  # logvar

        # Decoder for preference data x
        ae_structure_z = [z_dim + h_dim] + ae_structure_z[::-1]
        self.decoder = nn.Sequential()
        for i in range(len(ae_structure_z) - 1):
            self.decoder.add_module(
                "fc{}".format(i), nn.Linear(ae_structure_z[i], ae_structure_z[i + 1])
            )
            if i != len(ae_structure_z) - 2:
                self.decoder.add_module("act{}".format(i), self.act_fn)

    def encode_qz(self, y):
        o = self.encoder_qz(y)
        return self.enc_qz_mu(o), self.enc_qz_logvar(o)

    def encode_qhx(self, x):
        o = self.encoder_qhx(x)
        return self.enc_qhx_mu(o), self.enc_qhx_logvar(o)

    def encode_qhy(self, y):
        o = self.encoder_qhy(y)
        return self.enc_qhy_mu(o), self.enc_qhy_logvar(o)

    def encode_phx(self, x):
        o = self.encoder_phx(x)
        return self.enc_ph_mu(o), self.enc_ph_logvar(o)

    def decode(self, z, h):
        zh = torch.cat([z, h], 1)
        h = self.decoder(zh)
        if self.likelihood == "mult":
            return torch.softmax(h, dim=1)
        else:
            return torch.sigmoid(h)

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(mu)
        return mu + eps * std

    def forward(self, y, x):
        mu_qz, logvar_qz = self.encode_qz(y)
        mu_qhx, logvar_qhx = self.encode_qhx(x)
        mu_qhy, logvar_qhy = self.encode_qhy(y)
        mu_ph, logvar_ph = self.encode_phx(x)

        z = self.reparameterize(mu_qz, logvar_qz)
        h_q = self.reparameterize(mu_qhx, logvar_qhx)

        return self.decode(z, h_q), mu_qz, logvar_qz, mu_qhx, logvar_qhx, mu_qhy, logvar_qhy, mu_ph, logvar_ph

    def loss(self, x, x_, mu_qz, logvar_qz, mu_qhx, logvar_qhx, mu_qhy, logvar_qhy, mu_ph, logvar_ph, beta, alpha_1,
             alpha_2):
        # Likelihood
        ll_choices = {
            "mult": x * torch.log(x_ + EPS),
            "bern": x * torch.log(x_ + EPS) + (1 - x) * torch.log(1 - x_ + EPS),
            "gaus": -(x - x_) ** 2,
            "pois": x * torch.log(x_ + EPS) - x_,
        }

        ll = ll_choices.get(self.likelihood, None)
        if ll is None:
            raise ValueError("Supported likelihoods: {}".format(ll_choices.keys()))

        ll = torch.sum(ll, dim=1)

        # KL term z
        std_qz = torch.exp(0.5 * logvar_qz)
        kld_z = -0.5 * (1 + 2.0 * torch.log(std_qz) - mu_qz.pow(2) - std_qz.pow(2))
        kld_z = torch.sum(kld_z, dim=1)

        # KL term h
        std_qhx = torch.exp(0.5 * logvar_qhx)
        std_qhy = torch.exp(0.5 * logvar_qhy)
        std_ph = torch.exp(0.5 * logvar_ph)

        # KL(q(h|x)||p(h|x))
        kld_hx = -0.5 * (1 + 2.0 * torch.log(std_qhx) - (mu_qhx - mu_ph).pow(2) - std_qhx.pow(
            2))  # assuming std_ph is 1 for now
        kld_hx = torch.sum(kld_hx, dim=1)

        # KL(q(h|x)||q(h|y))
        kld_hy = -0.5 * (1 + 2.0 * torch.log(std_qhx) - 2.0 * torch.log(std_qhy) - (
                    (mu_qhx - mu_qhy).pow(2) + std_qhx.pow(2)) / std_qhy.pow(2))  # assuming std_ph is 1 for now
        kld_hy = torch.sum(kld_hy, dim=1)

        return torch.mean(beta * kld_z + alpha_1 * kld_hx + alpha_2 * kld_hy - ll)


def learn(
        cvae,
        train_set,
        n_epochs,
        batch_size,
        learn_rate,
        beta,
        alpha_1,
        alpha_2,
        verbose,
        device=torch.device("cpu"),
):
    optimizer = torch.optim.Adam(params=cvae.parameters(), lr=learn_rate)

    x = train_set.user_graph.matrix[: train_set.num_users, : train_set.num_users]
    y = train_set.matrix
    progress_bar = trange(1, n_epochs + 1, disable=not verbose)
    for _ in progress_bar:
        sum_loss = 0.0
        count = 0
        for batch_id, u_ids in enumerate(
                train_set.user_iter(batch_size, shuffle=False)
        ):
            y_batch = y[u_ids, :]
            y_batch.data = np.ones(len(y_batch.data))  # Binarize data
            y_batch = y_batch.A
            y_batch = torch.tensor(y_batch, dtype=torch.float32, device=device)

            x_batch = x[u_ids, :]
            x_batch = x_batch.A
            x_batch = torch.tensor(x_batch, dtype=torch.float32, device=device)

            # Reconstructed batch
            y_batch_, mu_qz, logvar_qz, mu_qhx, logvar_qhx, mu_qhy, logvar_qhy, mu_ph, logvar_ph = cvae(y_batch,
                                                                                                        x_batch)

            loss = cvae.loss(y_batch, y_batch_, mu_qz, logvar_qz, mu_qhx, logvar_qhx, mu_qhy, logvar_qhy, mu_ph,
                             logvar_ph, alpha_1, alpha_2, beta)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            sum_loss += loss.data.item()
            count += y_batch.shape[0]

            if batch_id % 10 == 0:
                progress_bar.set_postfix(loss=(sum_loss / count))

    return cvae



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from tqdm.auto import trange

from ..recommender import Recommender
from ...exception import ScoreException
from ...utils import get_rng
from ...utils.init_utils import xavier_uniform


class CVAE(Recommender):
    """
    Collaborative Variational Autoencoder

    Parameters
    ----------
    z_dim: int, optional, default: 50
        The dimension of the user and item latent factors.

    n_epochs: int, optional, default: 100
        Maximum number of epochs for training.

    lambda_u: float, optional, default: 1e-4
        The regularization hyper-parameter for user latent factor.

    lambda_v: float, optional, default: 0.001
        The regularization hyper-parameter for item latent factor.

    lambda_r: float, optional, default: 10.0
        Parameter that balance the focus on content or ratings

    lambda_w: float, optional, default: 1e-4
        The regularization for VAE weights

    lr: float, optional, default: 0.001
        Learning rate in the auto-encoder training

    a: float, optional, default: 1
        The confidence of observed ratings.

    b: float, optional, default: 0.01
        The confidence of unseen ratings.

    input_dim: int, optional, default: 8000
        The size of input vector

    vae_layers: list, optional, default: [200, 100]
        The list containing size of each layers in neural network structure

    act_fn: str, default: 'sigmoid'
        Name of the activation function used for the variational auto-encoder.
        Supported functions: ['sigmoid', 'tanh', 'elu', 'relu', 'relu6', 'leaky_relu', 'identity']

    loss_type: String, optional, default: "cross-entropy"
        Either "cross-entropy" or "rmse"
        The type of loss function in the last layer

    batch_size: int, optional, default: 128
        The batch size for SGD.

    init_params: dict, optional, default: {'U':None, 'V':None}
        Initial U and V latent matrix

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model already \
        pre-trained (U and V are not None).

    References
    ----------
    Collaborative Variational Autoencoder for Recommender Systems
    X. Li and J. She ACM SIGKDD International Conference on Knowledge Discovery and Data Mining 2017

    http://eelxpeng.github.io/assets/paper/Collaborative_Variational_Autoencoder.pdf

    """

    def __init__(
        self,
        name="CVAE",
        z_dim=50,
        n_epochs=100,
        lambda_u=1e-4,
        lambda_v=0.001,
        lambda_r=10,
        lambda_w=1e-4,
        lr=0.001,
        a=1,
        b=0.01,
        input_dim=8000,
        vae_layers=[200, 100],
        act_fn="sigmoid",
        loss_type="cross-entropy",
        batch_size=128,
        init_params=None,
        trainable=True,
        seed=None,
        verbose=True,
    ):
        super().__init__(name=name, trainable=trainable, verbose=verbose)

        self.lambda_u = lambda_u
        self.lambda_v = lambda_v
        self.lambda_r = lambda_r
        self.lambda_w = lambda_w
        self.a = a
        self.b = b
        self.n_epochs = n_epochs
        self.input_dim = input_dim
        self.vae_layers = vae_layers
        self.z_dim = z_dim
        self.loss_type = loss_type
        self.act_fn = act_fn
        self.lr = lr
        self.batch_size = batch_size
        self.seed = seed

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.U = self.init_params.get("U", None)
        self.V = self.init_params.get("V", None)

    def _init(self):
        rng = get_rng(self.seed)
        n_users, n_items = self.train_set.num_users, self.train_set.num_items

        if self.U is None:
            self.U = xavier_uniform((n_users, self.z_dim), rng)
        if self.V is None:
            self.V = xavier_uniform((n_items, self.z_dim), rng)

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        self._init()
        
        if self.trainable:
            self._fit_cvae()

        return self

    def _fit_cvae(self):
        R = self.train_set.csc_matrix  # csc for efficient slicing over items
        document = self.train_set.item_text.batch_bow(
            np.arange(self.train_set.num_items)
        )  # bag of word feature
        document = (document - document.min()) / (
            document.max() - document.min()
        )  # normalization

        # VAE initialization
        from .cvae import Model
        import tensorflow.compat.v1 as tf
        
        tf.disable_eager_execution()

        tf.set_random_seed(self.seed)
        model = Model(
            n_users=self.train_set.num_users,
            n_items=self.train_set.num_items,
            input_dim=self.input_dim,
            U=self.U,
            V=self.V,
            n_z=self.z_dim,
            lambda_u=self.lambda_u,
            lambda_v=self.lambda_v,
            lambda_r=self.lambda_r,
            lambda_w=self.lambda_w,
            layers=self.vae_layers,
            loss_type=self.loss_type,
            act_fn=self.act_fn,
            seed=self.seed,
            lr=self.lr,
        )

        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        sess = tf.Session(config=config)
        sess.run(tf.global_variables_initializer())  # init variable

        loop = trange(self.n_epochs, disable=not self.verbose)
        for _ in loop:
            cf_loss, vae_loss, count = 0, 0, 0
            for i, batch_ids in enumerate(
                self.train_set.item_iter(self.batch_size, shuffle=True)
            ):
                batch_R = R[:, batch_ids]
                batch_C = np.ones(batch_R.shape) * self.b
                batch_C[batch_R.nonzero()] = self.a

                feed_dict = {
                    model.x: document[batch_ids],
                    model.ratings: batch_R.A,
                    model.C: batch_C,
                    model.item_ids: batch_ids,
                }
                _, _vae_los = sess.run([model.vae_update, model.vae_loss], feed_dict)
                _, _cf_loss = sess.run([model.cf_update, model.cf_loss], feed_dict)

                cf_loss += _cf_loss
                vae_loss += _vae_los
                count += len(batch_ids)
                if i % 10 == 0:
                    loop.set_postfix(
                        vae_loss=(vae_loss / count), cf_loss=(cf_loss / count)
                    )

        self.U, self.V = sess.run([model.U, model.V])

        tf.reset_default_graph()

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            known_item_scores = self.V.dot(self.U[user_idx, :])
            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )

            user_pred = self.V[item_idx, :].dot(self.U[user_idx, :])

            return user_pred



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
import scipy.sparse as sp
import scipy as sc
import tensorflow as tf


class PCRL_:
    def __init__(
        self,
        train_set,
        k=100,
        z_dims=[300],
        n_epoch=300,
        batch_size=300,
        learning_rate=0.001,
        B=1,
        w_determinist=True,
        init_params=None,
    ):

        self.train_set = train_set
        self.cf_data = sp.csc_matrix(
            self.train_set.matrix
        )  # user-item interaction (CF data)
        self.aux_data = self.train_set.item_graph.matrix[
            : self.train_set.num_items, : self.train_set.num_items
        ]  # item auxiliary information (items'context in the original paper)
        self.k = k  # the number of user and item latent factors
        self.z_dims = (
            z_dims
        )  # the dimension of the second hidden layer (we consider a 2-layers PCRL)
        self.c_dim = self.aux_data.shape[
            1
        ]  # the dimension of the auxiliary information matrix
        self.n_epoch = n_epoch  # the number of traning epochs
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.B = B  # Shape augmentation parameter
        self.init_params = init_params  # initial parameters
        # Additional parameters
        self.aa = 0.3
        self.bb = 0.3
        self.Ls = sp.csc_matrix(
            (self.aux_data.shape[0], self.k)
        )  # Variational Shape parameters of the item factors (Beta in the paper)
        self.Lr = sp.csc_matrix(
            (self.aux_data.shape[0], self.k)
        )  # Variational Rate parameters  of the item factors (Beta in the paper)
        self.Gs = (
            None
        )  # Variational Shapre parameters of the user factors (Theta in the paper)
        self.Gr = (
            None
        )  # Variational Rate parameters of the user factors (Theta in the paper)
        self.L = len(z_dims)  # The number of deterministic hidden layers "z"
        self.w_determinist = (
            w_determinist
        )  # If true then deterministic wheights are used for the generator network
        self.sess = tf.Session()  # Tensorflow session
        # Inference netwok parameters
        self.inference_params = []
        self.inference_params.append(
            tf.Variable(self.glorot_init([self.c_dim, self.z_dims[self.L - 1]]))
        )
        for l in range(self.L - 2, -1, -1):
            self.inference_params.append(
                tf.Variable(self.glorot_init([self.z_dims[l + 1], self.z_dims[l]]))
            )
        self.inference_params.append(
            tf.Variable(self.glorot_init([self.z_dims[0], self.k]))
        )
        self.inference_params.append(
            tf.Variable(self.glorot_init([self.z_dims[0], self.k]))
        )
        # generator newtork parameters
        self.generator_params = []
        self.generator_params.append(
            tf.Variable(self.glorot_init([self.k, self.z_dims[0]]))
        )
        for l in range(1, self.L):
            self.generator_params.append(
                tf.Variable(self.glorot_init([self.z_dims[l - 1], self.z_dims[l]]))
            )
        self.generator_params.append(
            tf.Variable(self.glorot_init([self.z_dims[self.L - 1], self.c_dim]))
        )

    def glorot_init(self, shape):
        return tf.random_normal(shape=shape, stddev=1.0 / tf.sqrt(shape[0] / 2.0))

    # some until function to compute the loss
    # Log density of Ga(alpha, beta)
    def log_q(self, z, alpha, beta):
        return (
            (alpha - 1) * tf.log(z) - beta * z + alpha * tf.log(beta) - tf.lgamma(alpha)
        )

        # Log density of the standard normal N(0, 1)

    def log_t(self, epsilon):
        return -0.5 * tf.log(2 * np.pi) - 0.5 * epsilon ** 2

    # Marsaglia and Tsang transformation
    def G(self, epsilon, alpha, beta):
        return (
            (alpha - 1.0 / 3.0) * (1 + epsilon / tf.sqrt(9.0 * alpha - 3.0)) ** 3 / beta
        )

    # derivative of h
    def dG(self, epsilon, alpha, beta):
        return (
            (alpha - 1.0 / 3)
            * (3.0 / tf.sqrt(9.0 * alpha - 3.0))
            * (1.0 + epsilon / tf.sqrt(9.0 * alpha - 3.0)) ** 2
        ) / beta

    # Log density of the proposal distribution r(z) = t(epsilon) * |dG/depsilon|^{-1}
    def log_r(self, epsilon, alpha, beta):
        return -tf.log(self.dG(epsilon, alpha, beta)) + self.log_t(epsilon)

    # Inverse transformation
    def G_inv(self, z, alpha, beta):
        return tf.sqrt(9.0 * alpha - 3.0) * (
            (beta * z / (alpha - 1.0 / 3.0)) ** (1.0 / 3.0) - 1.0
        )

    # Sample from the marginal of the accepted epsilon's, epsilon ~ pi(epsilon)
    def sample_pi(self, alpha, beta):
        Gam = tf.random_gamma([1], alpha=alpha, beta=beta, name="Gam", seed=None)[0]
        return self.G_inv(Gam, alpha, beta)

    # shape augmentation
    def shape_augmentation(self, alpha, B):
        UUU = 1.0
        for i in range(int(B)):
            UUU = UUU * tf.pow(
                tf.random_uniform(tf.shape(alpha), seed=None), 1.0 / (alpha + i)
            )
        return UUU

    # Collaborative filtering part of pcrl (Poisson Factorization)
    def pf_(self, X, k, max_iter=1, init_params=None):

        # data preparation
        X = sp.csc_matrix(X, dtype=np.float64)
        M = X.copy()
        M.data = np.full(len(M.data), 1)
        n = X.shape[0]
        d = X.shape[1]
        eps = 0.000000001

        # Hyper-parameters setting
        a = 0.3

        # Parameter initialization

        # shape gamma_uk matrix (dgCMatrix)
        if init_params["G_s"] is None:
            G_s = np.random.gamma(50, scale=0.3 / 50, size=n * k).reshape(n, k)
        else:
            G_s = init_params["G_s"]
        G_s = sp.csc_matrix(G_s, dtype=np.float64)

        ## rate gamma_uk matrix (dgCMatrix)
        if init_params["G_r"] is None:
            G_r = np.random.gamma(50, scale=0.3 / 50, size=n * k).reshape(n, k)
        else:
            G_r = init_params["G_r"]
        G_r = sp.csc_matrix(G_r, dtype=np.float64)

        # shape lamda_ik matrix (dgCMatrix)
        if init_params["L_s"] is None:
            L_s = np.random.gamma(50, scale=0.3 / 50, size=d * k).reshape(d, k)
        else:
            L_s = init_params["L_s"]
        L_s = sp.csc_matrix(L_s, dtype=np.float64)

        ## rate lamda_ik matrix (dgCMatrix)
        if init_params["L_r"] is None:
            L_r = np.random.gamma(50, scale=0.3 / 50, size=d * k).reshape(d, k)
        else:
            L_r = init_params["L_r"]
        L_r = sp.csc_matrix(L_r, dtype=np.float64)

        # need to be computed only once as Lr and Ls don't change here
        logL_r = L_r.copy()
        logL_r.data = np.log(logL_r.data)
        digaL_s = L_s.copy()
        digaL_s.data = sc.special.digamma(digaL_s.data)

        Lb = digaL_s - logL_r
        Lb.data = np.exp(Lb.data)

        del logL_r
        del digaL_s

        Lb = Lb.todense()

        # Learning
        for iter_ in range(1, max_iter + 1):
            ## Update multinomiale parameter, no need to store phi only compute the sufficient stats
            logG_r = G_r.copy()
            logG_r.data = np.log(logG_r.data)
            digaG_s = G_s.copy()
            digaG_s.data = sc.special.digamma(digaG_s.data)

            Lt = digaG_s - logG_r
            Lt.data = np.exp(Lt.data)

            del logG_r
            del digaG_s

            Lt = Lt.todense()

            ## Update user related parameters
            G_s = a + np.multiply(Lt, ((X / (Lt * Lb.T + eps)) * Lb))
            G_r = np.repeat(np.sum(L_s / L_r, 0), n, axis=0) + a

            G_s = sp.csc_matrix(G_s)
            G_r = sp.csc_matrix(G_r)

        Tk = np.repeat(np.sum(G_s / G_r, 0), self.batch_size, axis=0)
        Zik = np.multiply(Lb, ((X.T / (Lb * Lt.T + eps)) * Lt))
        # End of learning

        res = {
            "Z": G_s / G_r,
            "W": L_s / L_r,
            "G_s": G_s,
            "G_r": G_r,
            "L_s": L_s,
            "L_r": L_r,
            "Lt": Lt,
            "Lb": Lb,
            "Zik": np.array(Zik, dtype="float32"),
            "Tk": np.array(Tk, dtype="float32"),
        }

        return res

    # The inference network (or encoder)
    def inference_net(self, C, reuse=None):
        # input
        # h = tf.nn.relu(tf.sparse_matmul(C, self.inference_params[0], a_is_sparse=True))
        h = tf.nn.relu(tf.matmul(C, self.inference_params[0]))
        # intermediate hidden layer
        for l in range(1, self.L):
            # h = tf.nn.relu(tf.sparse_matmul(h, self.inference_params[l], a_is_sparse=True))
            h = tf.nn.relu(tf.matmul(h, self.inference_params[l]))

        # output
        # beta = tf.nn.softplus(tf.matmul(h, self.inference_params[self.L], a_is_sparse=True)) + 0.3
        # alpha = tf.nn.softplus(tf.matmul(h, self.inference_params[self.L + 1], a_is_sparse=True)) + 0.3
        beta = tf.nn.softplus(tf.matmul(h, self.inference_params[self.L])) + 0.3
        alpha = tf.nn.softplus(tf.matmul(h, self.inference_params[self.L + 1])) + 0.3

        return alpha, beta

    # The generative network (or decoder)
    def generative_net(self, Z, reuse=None):

        # with tf.variable_scope("generative",reuse=reuse):
        if self.w_determinist:
            h2 = tf.nn.relu(tf.matmul(Z, self.generator_params[0]))
            for l in range(1, self.L):
                # h2 = tf.nn.relu(tf.sparse_matmul(h2, self.generator_params[l], a_is_sparse=True))
                h2 = tf.nn.relu(tf.matmul(h2, self.generator_params[l]))
            # d_x = tf.nn.sigmoid(tf.matmul(h2, self.generator_params[self.L], a_is_sparse=True))
            d_x = tf.nn.sigmoid(tf.matmul(h2, self.generator_params[self.L]))
        else:
            e = tf.random_normal(
                tf.shape(self.generator_params[0]),
                dtype=tf.float32,
                mean=0.0,
                stddev=1.0,
                name="epsilon",
            )
            h2 = tf.nn.relu(tf.matmul(Z, self.generator_params[0] + 0.01 * e))
            for l in range(1, self.L):
                e = tf.random_normal(
                    tf.shape(self.generator_params[l]),
                    dtype=tf.float32,
                    mean=0.0,
                    stddev=1.0,
                    name="epsilon",
                )
                # h2 = tf.nn.relu(tf.sparse_matmul(h2, self.generator_params[l] + 0.01 * e, a_is_sparse=True))
                h2 = tf.nn.relu(tf.matmul(h2, self.generator_params[l] + 0.01 * e))
            e = tf.random_normal(
                tf.shape(self.generator_params[self.L]),
                dtype=tf.float32,
                mean=0.0,
                stddev=1.0,
                name="epsilon",
            )
            # d_x = tf.nn.sigmoid(tf.matmul(h2, self.generator_params[self.L] + 0.01 * e, a_is_sparse=True))
            d_x = tf.nn.sigmoid(tf.matmul(h2, self.generator_params[self.L] + 0.01 * e))
        return d_x

        # The loss function

    def loss(self, C, X_g, X_, alpha, beta, z, E, Zik, Tk):

        const_term = C * tf.log(1e-10 + X_) - X_
        const_term = tf.reduce_sum(const_term, 1)

        loss1 = C * tf.log(1e-10 + X_g) - X_g
        loss1 = tf.reduce_sum(loss1, 1)

        loss2 = self.log_q(z, alpha + self.B, beta)
        loss2 = const_term * tf.reduce_sum(loss2, 1)

        # loss3 = -log_r(E, alpha,beta)
        loss3 = -self.log_r(E, alpha + self.B, beta)
        loss3 = const_term * tf.reduce_sum(loss3, 1)

        # The sum of KL terms of all generator's wheights (up to constant terms)
        kl_w = 0.0
        if not self.w_determinist:
            for l in range(0, self.L + 1):
                kl_w += tf.reduce_sum(
                    -0.5 * tf.reduce_sum(tf.square(self.generator_params[l]), 1)
                )

        # KL Divergence term
        kl_term = (
            (alpha - self.aa - Zik) * tf.digamma(alpha)
            - tf.lgamma(alpha)
            + (self.aa + Zik) * tf.log(beta)
            + alpha * (Tk + self.bb - beta) / beta
        )
        kl_term = -tf.reduce_sum(kl_term, 1)
        return (
            -tf.reduce_mean(loss1 + loss2 + loss3 + kl_term)
            + kl_w / self.aux_data.shape[0]
        )

    # fitting PCRL to observed data
    def learn(self):
        # placeholders
        C = tf.placeholder(tf.float32, shape=[None, self.c_dim], name="C")
        X_ = tf.placeholder(tf.float32, shape=[None, self.c_dim], name="X_")
        Zik = tf.placeholder(tf.float32, shape=[None, self.k], name="Zik")
        Tk = tf.placeholder(tf.float32, shape=[None, self.k], name="Tk")
        E = tf.placeholder(
            tf.float32, shape=[None, self.k], name="E"
        )  # matrix of accepted samples epsilon, epsilon ~ pi(epsilon)

        # Sample Gamma variables using the Reparameterized Acceptance-Rejection approach
        alpha, beta = self.inference_net(C)
        E_ = self.sample_pi(alpha + self.B, beta)
        z_tld = self.G(E, alpha + self.B, beta)
        # getting the right z, this step is necessary when alpha<1.0
        U_ = self.shape_augmentation(alpha, self.B)
        z = U_ * z_tld

        # Generating
        X_g = self.generative_net(z)

        # preparing optimization
        loss = self.loss(C, X_g, X_, alpha, beta, z_tld, E, Zik, Tk)
        # train = tf.train.AdamOptimizer(learning_rate=learning_rate).minimize(loss)
        train = tf.train.RMSPropOptimizer(learning_rate=self.learning_rate).minimize(
            loss
        )

        # Initialization
        init = tf.global_variables_initializer()

        # Learning
        self.sess.run(init)

        # Train the collaborative part
        resPF = self.pf_(
            self.cf_data, k=self.k, max_iter=1, init_params=self.init_params
        )

        for epoch in range(self.n_epoch):
            for idx in self.train_set.item_iter(self.batch_size, shuffle=False):
                batch_C = self.aux_data[idx].A
                EE = self.sess.run(E_, feed_dict={C: batch_C})
                z_c = self.sess.run(X_g, feed_dict={C: batch_C, E: EE})
                feed_dict = {
                    C: batch_C,
                    X_: z_c,
                    E: EE,
                    Zik: resPF["Zik"][idx],
                    Tk: resPF["Tk"][0 : len(idx)],
                }
                _, l = self.sess.run([train, loss], feed_dict=feed_dict)
                del (EE, z_c)
            for idx in self.train_set.item_iter(2 * self.batch_size, shuffle=False):
                batch_C = self.aux_data[idx].A
                self.Ls[idx], self.Lr[idx] = self.sess.run(
                    [alpha, beta], feed_dict={C: batch_C}
                )
            print("epoch %i, Train Loss: %f" % (epoch, l))
            resPF = self.pf_(
                self.cf_data,
                k=self.k,
                max_iter=1,
                init_params={
                    "G_s": resPF["G_s"],
                    "G_r": resPF["G_r"],
                    "L_s": self.Ls,
                    "L_r": self.Lr,
                },
            )
        self.Gs = resPF["G_s"]
        self.Gr = resPF["G_r"]
        print("learning done successfully")
        # End of traning

        return self



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from ..recommender import Recommender


class PCRL(Recommender):
    """Probabilistic Collaborative Representation Learning.

    Parameters
    ----------
    k: int, optional, default: 100
        The dimension of the latent factors.
        
    z_dims: Numpy 1d array, optional, default: [300]
        The dimensions of the hidden intermdiate layers 'z' in the order \
        [dim(z_L), ...,dim(z_1)], please refer to Figure 1 in the orginal paper for more details.

    max_iter: int, optional, default: 300
        Maximum number of iterations (number of epochs) for variational PCRL.
        
    batch_size: int, optional, default: 300
        The batch size for SGD.
        
    learning_rate: float, optional, default: 0.001
        The learning rate for SGD.

    aux_info: see "cornac/examples/pcrl_example.py" in the GitHub repo for an example of how to use \
        cornac's graph modality provide item auxiliary data (e.g., context, text, etc.) for PCRL.

    name: string, optional, default: 'PCRL'
        The name of the recommender model.

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model already \
        pre-trained (Theta, Beta and Xi are not None). 
        
    w_determinist: boolean, optional, default: True
        When True, determinist wheights "W" are used for the generator network, \
        otherwise "W" is stochastic as in the original paper.

    init_params: dictionary, optional, default: None
        List of initial parameters, e.g., init_params = {'G_s':G_s, 'G_r':G_r, 'L_s':L_s, 'L_r':L_r}.
        
        Theta: ndarray, shape (n_users, k)
            The expected user latent factors.

        Beta: ndarray, shape (n_items, k)
            The expected item latent factors.

        G_s: ndarray, shape (n_users, k)
            Represent the "shape" parameters of Gamma distribution over Theta.
            
        G_r: ndarray, shape (n_users, k)
            Represent the "rate" parameters of Gamma distribution over Theta. 
        
        L_s: ndarray, shape (n_items, k)
            Represent the "shape" parameters of Gamma distribution over Beta.
            
        L_r: ndarray, shape (n_items, k)
            Represent the "rate" parameters of Gamma distribution over Beta. 
        
    References
    ----------
    * Salah, Aghiles, and Hady W. Lauw. Probabilistic Collaborative Representation Learning for Personalized Item Recommendation. \
    In UAI 2018.
    """

    def __init__(
        self,
        k=100,
        z_dims=[300],
        max_iter=300,
        batch_size=300,
        learning_rate=0.001,
        name="PCRL",
        trainable=True,
        verbose=False,
        w_determinist=True,
        init_params=None,
    ):

        Recommender.__init__(self, name=name, trainable=trainable, verbose=verbose)

        self.k = k
        self.z_dims = (
            z_dims
        )  # the dimension of the second hidden layer (we consider a 2-layers PCRL)
        self.max_iter = max_iter
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.w_determinist = w_determinist

        # Init params if provided
        self.init_params = {} if init_params is None else init_params
        self.Theta = self.init_params.get("Theta", None)
        self.Beta = self.init_params.get("Beta", None)
        self.Gs = self.init_params.get("G_s", None)
        self.Gr = self.init_params.get("G_r", None)
        self.Ls = self.init_params.get("L_s", None)
        self.Lr = self.init_params.get("L_r", None)

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        # X = sp.csc_matrix(self.train_set.matrix)

        if self.trainable:
            from .pcrl import PCRL_

            # use pre-trained params if exists, otherwise from constructor
            init_params = {
                "G_s": self.Gs,
                "G_r": self.Gr,
                "L_s": self.Ls,
                "L_r": self.Lr,
            }

            # instanciate pcrl
            # train_aux_info = train_set.item_graph.matrix[:self.train_set.num_items, :self.train_set.num_items]
            pcrl_ = PCRL_(
                train_set=train_set,
                k=self.k,
                z_dims=self.z_dims,
                n_epoch=self.max_iter,
                batch_size=self.batch_size,
                learning_rate=self.learning_rate,
                B=1,
                w_determinist=self.w_determinist,
                init_params=init_params,
            ).learn()

            self.Theta = np.array(pcrl_.Gs) / np.array(pcrl_.Gr)
            self.Beta = np.array(pcrl_.Ls) / np.array(pcrl_.Lr)

            # overwrite init_params for future fine-tuning
            self.Gs = pcrl_.Gs
            self.Gr = pcrl_.Gr
            self.Ls = pcrl_.Ls
            self.Lr = pcrl_.Lr

        elif self.verbose:
            print("%s is trained already (trainable = False)" % (self.name))

        return self

    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for a list of items.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """

        if item_idx is None:
            user_pred = self.Beta * self.Theta[user_idx, :].T
        else:
            user_pred = self.Beta[item_idx, :] * self.Theta[user_idx, :].T
        # transform user_pred to a flatten array
        user_pred = np.array(user_pred, dtype="float64").flatten()

        return user_pred



import numpy as np

from cornac.models.recommender import Recommender
from cornac.exception import ScoreException

class EASE(Recommender):
    """Embarrassingly Shallow Autoencoders for Sparse Data.

    Parameters
    ----------
    name: string, optional, default: 'EASEᴿ'
        The name of the recommender model.

    lamb: float, optional, default: 500
        L2-norm regularization-parameter λ ∈ R+.
    
    posB: boolean, optional, default: False
        Remove Negative Weights

    trainable: boolean, optional, default: True
        When False, the model is not trained and Cornac assumes that the model is already \
        trained. 
        
    verbose: boolean, optional, default: False
        When True, some running logs are displayed.

    seed: int, optional, default: None
        Random seed for parameters initialization.

    References
    ----------
    * Steck, H. (2019, May). "Embarrassingly shallow autoencoders for sparse data." \
    In The World Wide Web Conference (pp. 3251-3257).
    """

    def __init__(
            self,
            name="EASEᴿ",
            lamb=500,
            posB=True,
            trainable=True,
            verbose=True,
            seed=None,
            B=None,
            U=None,
    ):
        Recommender.__init__(self, name=name, trainable=trainable, verbose=verbose)
        self.lamb = lamb
        self.posB = posB
        self.verbose = verbose
        self.seed = seed
        self.B = B
        self.U = U

    def fit(self, train_set, val_set=None):
        """Fit the model to observations.

        Parameters
        ----------
        train_set: :obj:`cornac.data.Dataset`, required
            User-Item preference data as well as additional modalities.

        val_set: :obj:`cornac.data.Dataset`, optional, default: None
            User-Item preference data for model selection purposes (e.g., early stopping).

        Returns
        -------
        self : object
        """
        Recommender.fit(self, train_set, val_set)

        # A rating matrix
        self.U = self.train_set.matrix

        # Gram matrix is X^t X, compute dot product
        G = self.U.T.dot(self.U).toarray()

        diag_indices = np.diag_indices(G.shape[0])

        G[diag_indices] = G.diagonal() + self.lamb

        P = np.linalg.inv(G)

        B = P / (-np.diag(P))
        
        B[diag_indices] = 0.0

        # if self.posB remove -ve values
        if self.posB:
            B[B<0]=0            

        # save B for predictions
        self.B=B
      
        return self


    def score(self, user_idx, item_idx=None):
        """Predict the scores/ratings of a user for an item.

        Parameters
        ----------
        user_idx: int, required
            The index of the user for whom to perform score prediction.

        item_idx: int, optional, default: None
            The index of the item for which to perform score prediction.
            If None, scores for all known items will be returned.

        Returns
        -------
        res : A scalar or a Numpy array
            Relative scores that the user gives to the item or to all known items

        """
        if item_idx is None:
            if self.train_set.is_unk_user(user_idx):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d)" % user_idx
                )

            known_item_scores = self.U[user_idx, :].dot(self.B)
            return known_item_scores
        else:
            if self.train_set.is_unk_user(user_idx) or self.train_set.is_unk_item(
                item_idx
            ):
                raise ScoreException(
                    "Can't make score prediction for (user_id=%d, item_id=%d)"
                    % (user_idx, item_idx)
                )

            user_pred = self.B[item_idx, :].dot(self.U[user_idx, :])

            return user_pred


# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from .recommender import Recommender

from .amr import AMR
from .baseline_only import BaselineOnly
from .bivaecf import BiVAECF
from .bpr import BPR
from .bpr import WBPR
from .causalrec import CausalRec
from .c2pf import C2PF
from .cdl import CDL
from .cdr import CDR
from .coe import COE
from .comparer import ComparERObj
from .comparer import ComparERSub
from .conv_mf import ConvMF
from .ctr import CTR
from .cvae import CVAE
from .cvaecf import CVAECF
from .ease import EASE
from .efm import EFM
from .global_avg import GlobalAvg
from .hft import HFT
from .hpf import HPF
from .hrdr import HRDR
from .ibpr import IBPR
from .knn import ItemKNN
from .knn import UserKNN
from .mcf import MCF
from .mf import MF
from .mmmf import MMMF
from .most_pop import MostPop
from .mter import MTER
from .narre import NARRE
from .ncf import GMF
from .ncf import MLP
from .ncf import NeuMF
from .nmf import NMF
from .online_ibpr import OnlineIBPR
from .pcrl import PCRL
from .pmf import PMF
from .sbpr import SBPR
from .skm import SKMeans
from .sorec import SoRec
from .svd import SVD
from .vaecf import VAECF
from .vbpr import VBPR
from .vmf import VMF
from .wmf import WMF

try:
    from .fm import FM
except ModuleNotFoundError:
    print(
        "FM model is only supported on Linux.\n"
        + "Windows executable can be found at http://www.libfm.org."
    )




# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import random

import numpy as np
import torch

"""Firstly, we define a helper function to generate\sample training ordinal triplets:
   Step 1:  
   given rated item i, randomly choose item j and check whether rating of j is lower than i, 
   if not randomly sample another item. 
   each row of the sampled data in the following form:
        [userId itemId_i itemId_j]
   for each user u, he/she prefers item i over item j.
"""


def sample_triplet(X, batch_size):
    sampled_data = np.zeros((batch_size, 3), dtype=np.int)

    count = 0
    while count < batch_size:
        u = random.randint(0, X.shape[0] - 1)
        u_row = X.getrow(u)
        _, u_nz = u_row.nonzero()
        min_rating = u_row[:, u_nz].todense().min()

        i = u_nz[random.randint(0, len(u_nz) - 1)]
        ratingi = u_row[:, i]

        if ratingi > min_rating:
            j = u_nz[random.randint(0, len(u_nz) - 1)]

            while u_row[:, j] >= ratingi:
                j = u_nz[random.randint(0, len(u_nz) - 1)]

            sampled_data[count, :] = [u, i, j]
            count += 1

    print("Done sampling")
    return sampled_data


def coe(
    X,
    k,
    lamda=0.05,
    n_epochs=150,
    learning_rate=0.001,
    batch_size=1000,
    init_params=None,
):
    # Data = Dataset(data)

    # Initial user factors
    if init_params["U"] is None:
        U = torch.randn(X.shape[0], k, requires_grad=True)
    else:
        U = torch.from_numpy(init_params["U"])
        U.requires_grad = True

    # Initial item factors
    if init_params["V"] is None:
        V = torch.randn(X.shape[1], k, requires_grad=True)
    else:
        V = torch.from_numpy(init_params["V"])
        V.requires_grad = True
        
    optimizer = torch.optim.Adam([U, V], lr=learning_rate)
    for epoch in range(n_epochs):
        #        num_steps = int(Data.data.shape[0]/batch_size)

        #       for i in range(1, num_steps + 1):
        #          batch_c,_ = Data.next_batch(batch_size)
        sampled_batch = sample_triplet(X, batch_size)

        regU = U[sampled_batch[:, 0], :]
        regI = V[sampled_batch[:, 1], :]
        regJ = V[sampled_batch[:, 2], :]

        regU_unq = U[np.unique(sampled_batch[:, 0]), :]
        regI_unq = V[np.unique(sampled_batch[:, 1:]), :]

        Scorei = torch.norm(regU - regI, dim=1)
        Scorej = torch.norm(regU - regJ, dim=1)

        loss = (
            lamda * (regU_unq.norm().pow(2) + regI_unq.norm().pow(2))
            - torch.log(torch.sigmoid(Scorej - Scorei)).sum()
        )
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        print("epoch:", epoch, "loss:", loss)

    U = U.data.cpu().numpy()
    V = V.data.cpu().numpy()

    res = {"U": U, "V": V}

    return res



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================


from tqdm.auto import tqdm
import numpy as np
import torch

from ...utils.common import scale
from ...utils.init_utils import normal
from ...utils import get_rng

torch.set_default_dtype(torch.double)


def _load_or_randn(size, init_values, seed, device):
    if init_values is None:
        rng = get_rng(seed)
        tensor = normal(size, mean=0.0, std=0.001, random_state=rng, dtype=np.double)
        tensor = torch.tensor(tensor, requires_grad=True, device=device)
    else:
        tensor = torch.tensor(init_values, requires_grad=True, device=device)
    return tensor


def _l2_loss(*tensors):
    l2_loss = 0
    for tensor in tensors:
        l2_loss += torch.sum(tensor ** 2) / 2
    return l2_loss


def vmf(
    train_set,
    item_feature,
    k,
    d,
    n_epochs,
    batch_size,
    lambda_u,
    lambda_v,
    lambda_p,
    lambda_e,
    learning_rate,
    gamma,
    init_params,
    use_gpu,
    verbose,
    seed,
):
    if use_gpu and torch.cuda.is_available():
        device = torch.device("cuda:0")
    else:
        device = torch.device("cpu")

    F = torch.from_numpy(item_feature).double().to(device)

    f_dim = train_set.item_image.feature_dim
    n_users = train_set.num_users
    n_items = train_set.num_items

    # preparing parameters
    U = _load_or_randn(
        (n_users, k), init_values=init_params.get("U"), seed=seed, device=device
    )
    V = _load_or_randn(
        (n_items, k), init_values=init_params.get("V"), seed=seed, device=device
    )
    P = _load_or_randn(
        (n_users, d), init_values=init_params.get("P"), seed=seed, device=device
    )
    E = _load_or_randn(
        (f_dim, d), init_values=init_params.get("E"), seed=seed, device=device
    )

    # optimizer
    optimizer = torch.optim.RMSprop([U, V, P, E], lr=learning_rate, alpha=gamma)

    for epoch in range(1, n_epochs + 1):
        sum_loss = 0.0
        count = 0
        progress_bar = tqdm(
            total=train_set.num_batches(batch_size),
            desc="Epoch {}/{}".format(epoch, n_epochs),
            disable=not verbose,
        )

        for batch_u, batch_i, batch_r in train_set.uir_iter(batch_size, shuffle=True):
            U_u = U[batch_u]
            P_u = P[batch_u]
            V_i = V[batch_i]
            f_i = F[batch_i]

            Rui = scale(batch_r, 0.0, 1.0, train_set.min_rating, train_set.max_rating)
            Rui = torch.tensor(Rui, device=device)

            Xui = torch.sigmoid(
                torch.sum(U_u * V_i, dim=1) + torch.sum(P_u * f_i.mm(E), dim=1)
            )

            loss = _l2_loss(Rui - Xui)
            reg = (
                lambda_u * _l2_loss(U_u)
                + lambda_v * _l2_loss(V_i)
                + lambda_p * _l2_loss(P_u)
                + lambda_e * _l2_loss(E)
            )
            loss = loss + reg

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            sum_loss += loss.data.item()
            count += len(batch_u)
            if count % (batch_size * 10) == 0:
                progress_bar.set_postfix(loss=(sum_loss / count))
            progress_bar.update(1)
        progress_bar.close()

        if verbose:
            print(sum_loss)

    res = {
        "U": U.data.cpu().numpy(),
        "V": V.data.cpu().numpy(),
        "P": P.data.cpu().numpy(),
        "E": E.data.cpu().numpy(),
        "Q": F.mm(E).data.cpu().numpy(),
    }

    return res



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from .modality import Modality
from .modality import FeatureModality
from .text import TextModality, ReviewModality
from .image import ImageModality
from .graph import GraphModality
from .sentiment import SentimentModality
from .reader import Reader
from .dataset import Dataset

__all__ = ['FeatureModality',
           'TextModality',
           'ReviewModality',
           'ImageModality',
           'GraphModality',
           'SentimentModality',
           'Dataset',
           'Reader']



# Copyright 2018 The Cornac Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np


class RatingMetric:
    """Rating Metric.

    Attributes
    ----------
    name: string,
        Name of the measure.

    type: string, value: 'rating'
        Type of the metric, e.g., "ranking", "rating".

    """

    def __init__(self, name=None, higher_better=False):
        self.type = 'rating'
        self.name = name
        self.higher_better = higher_better

    def compute(self, **kwargs):
        raise NotImplementedError()


class MAE(RatingMetric):
    """Mean Absolute Error.

    Attributes
    ----------
    name: string, value: 'MAE'
        Name of the measure.

    """

    def __init__(self):
        RatingMetric.__init__(self, name='MAE')

    def compute(self, gt_ratings, pd_ratings, weights=None, **kwargs):
        """Compute Mean Absolute Error.

        Parameters
        ----------
        gt_ratings: Numpy array
            Ground-truth rating values.

        pd_ratings: Numpy array
            Predicted rating values.

        weights: Numpy array, optional, default: None
            Weights for rating values.

        **kwargs: For compatibility

        Returns
        -------
        mae: A scalar.
            Mean Absolute Error.

        """
        mae = np.average(np.abs(gt_ratings - pd_ratings), axis=0, weights=weights)
        return mae


class MSE(RatingMetric):
    """Mean Squared Error.

    Attributes
    ----------
    name: string, value: 'MSE'
        Name of the measure.

    """

    def __init__(self):
        RatingMetric.__init__(self, name='MSE')

    def compute(self, gt_ratings, pd_ratings, weights=None, **kwargs):
        """Compute Mean Squared Error.

        Parameters
        ----------
        gt_ratings: Numpy array
            Ground-truth rating values.

        pd_ratings: Numpy array
            Predicted rating values.

        weights: Numpy array, optional, default: None
            Weights for rating values.

        **kwargs: For compatibility

        Returns
        -------
        mse: A scalar.
            Mean Squared Error.

        """
        mse = np.average((gt_ratings - pd_ratings) ** 2, axis=0, weights=weights)
        return mse


class RMSE(RatingMetric):
    """Root Mean Squared Error.

    Attributes
    ----------
    name: string, value: 'RMSE'
        Name of the measure.

    """

    def __init__(self):
        RatingMetric.__init__(self, name='RMSE')

    def compute(self, gt_ratings, pd_ratings, weights=None, **kwargs):
        """Compute Root Mean Squared Error.

        Parameters
        ----------
        gt_ratings: Numpy array
            Ground-truth rating values.

        pd_ratings: Numpy array
            Predicted rating values.

        weights: Numpy array, optional, default: None
            Weights for rating values.

        **kwargs: For compatibility

        Returns
        -------
        rmse: A scalar.
            Root Mean Squared Error.

        """
        mse = np.average((gt_ratings - pd_ratings) ** 2, axis=0, weights=weights)
        return np.sqrt(mse)
